from pathlib import (
    Path,
    PurePosixPath,
)
import pytest

from .. import UnsupportedRequest
from ..archivist import ArchivistRemote
from datalad_next.datasets import Dataset
from datalad_next.runners import CommandError

from datalad_next.tests import assert_result_count


@pytest.fixture(autouse=False, scope="function")
def archivist_dataset(tmp_path_factory, no_result_rendering):
    wpath = tmp_path_factory.mktemp("archivistds")
    # now create a second dataset that can pull all its content from
    # archives
    ads = Dataset(wpath / 'archiveds').create()
    # configure the archivist special remote for all dl_archive URL
    # handling
    ads.repo.call_annex([
        'initremote', 'archivist',
        'type=external', 'externaltype=archivist', 'encryption=none',
        'autoenable=true',
    ])
    return ads


@pytest.fixture(autouse=False, scope="function")
def populated_archivist_dataset(
        archivist_dataset, tmp_path_factory, no_result_rendering):
    """Returns a path to generated dataset

    This dataset references an annex archive with no other annexed files.
    The datalad special remote 'archivist' is enabled in the dataset and
    also set to autoenable.

    Returns
    -------
    Dataset, str, str, tuple(tuples)
      1. generated dataset instance
      2. the annex key for the included archive
      3. the leading directory of all files in the archive
      4. iterable with POSIX-path:content pairs for archive members.
         The path is relative to the leading archive directory, and
         can also be interpreted relative to the dataset root.
    """
    wpath = tmp_path_factory.mktemp("archivistds")

    ads = archivist_dataset

    dscontent = (
        ('azip/file1.txt', 'zipfile1'),
        ('azip/file2.csv', 'zipfile2_muchcontent'),
        ('atar/file1.txt', 'tarfile1'),
        ('atar/file2.csv', 'tarfile2_muchcontent'),
    )
    srcds = Dataset(wpath / 'srcds').create()
    for fpath, fcontent in dscontent:
        fpath = srcds.pathobj / (PurePosixPath(fpath))
        fpath.parent.mkdir(parents=True, exist_ok=True)
        fpath.write_text(fcontent)
    srcds.save()

    archive_root = wpath / 'myarchive'
    #archivetype = 'zip'

    akeys = {}

    for archivetype, ext in (('zip', ''), ('tar', '.gz')):
        archive_path = Path(f"{archive_root}.{archivetype}{ext}")

        archive_path_inds = ads.pathobj / '.archives' / archive_path.name
        # create an archive, the easy way, by simply exporting the
        # entire dataset worktree
        srcds.export_archive(archive_root, archivetype=archivetype)
        assert archive_path.exists()

        # add the archive (in a hidden dir) to be able to reference
        # it via a key
        aurl = archive_path.as_uri()
        ads.repo.call_annex([
            'addurl', '--file', str(archive_path_inds), aurl])
        ads.save()
        # get the key of the archive
        akeys[archivetype] = ads.status(
            archive_path_inds, annex='basic', return_type='item-or-list',
        )['key']
    return ads, akeys, archive_root, dscontent


def _check_archivist_addurl(atypes, ads, akeys, archive_root, dscontent):
    # run addurl on dl+archive URLs: this exercises CLAIMURL, CHECKURL,
    # TRANSFER_RETRIEVE
    for archivetype in atypes:
        for fpath, fcontent in dscontent:
            # take individual files from archives of different types
            if not fpath.startswith(f'a{archivetype}'):
                continue
            ads.repo.call_annex([
                'addurl',
                '--file',
                str(PurePosixPath(fpath)),
                f'dl+archive:{akeys[archivetype]}'
                f'#path={archive_root.name}/{fpath}&size={len(fcontent)}',
            ])
    # check that we reached the desired state
    whereis = ads.repo.whereis(
        # this dance is needed, because the normalize_paths decorator
        # of whereis() required platform paths
        str(Path('atar', 'file1.txt')),
        output='full',
    )
    # the file is known to exactly one remote (besides "here")
    assert len(whereis) == 2
    # and one remote is the archivist remote, and importantly not the
    # 'web' remote -- which would indicate a failure of claimurl/checkurl
    assert any(wi['description'] == '[archivist]' for wi in whereis.values())


def _check_archivist_retrieval(archivist_dataset):
    ads, akeys, archive_root, dscontent = archivist_dataset

    # step 1: addurl
    _check_archivist_addurl(
        # check all archive types supported
        ('zip', 'tar'),
        ads, akeys, archive_root, dscontent,
    )

    # make a clean dataset
    ads.save()

    # step 2: drop keys with dl+archive: URLs
    # now drop all archive member content. this should work,
    # because for each file there is a URL on record
    # -- hence always another copy
    # this requires archivist's CHECKPRESENT to function properly
    res = ads.drop(['azip', 'atar'])
    assert_result_count(
        res,
        # all files, plus the two directories we gave as arguments
        len(dscontent) + 2,
        action='drop',
        status='ok',
    )

    # step 3: retrieve keys with dl+archive: URLs from locally present archives
    res = ads.get(['azip', 'atar'])
    assert_result_count(
        res,
        len(dscontent),
        action='get',
        status='ok',
        type='file',
    )
    for fpath, fcontent in dscontent:
        assert (ads.pathobj / fpath).read_text() == fcontent

    # step 4: now drop ALL keys (incl. archives)
    # this will present a challenge for CHECKPRESENT:
    # without the archives no longer being around, it would requires remote
    # access or download to actually verify continued presence.
    # force this condition by dropping the archive keys first
    res = ads.drop('.archives')
    assert_result_count(
        res,
        # a tar and a zip
        2,
        action='drop',
        type='file',
        status='ok',
    )
    # and 4a now drop the keys that have their content from archives
    res = ads.drop(['azip', 'atar'])
    assert_result_count(
        res,
        len(dscontent),
        action='drop',
        status='ok',
        type='file',
    )
    # and now get again, this time with no archives around locally
    res = ads.get(['azip', 'atar'])
    assert_result_count(
        res,
        # no ZIP just yet
        len(dscontent),
        action='get',
        status='ok',
        type='file',
    )
    for fpath, fcontent in dscontent:
        assert (ads.pathobj / fpath).read_text() == fcontent
    # and drop everything again to leave the dataset empty
    res = ads.drop(['.'])


def test_archivist_retrieval(populated_archivist_dataset, no_result_rendering):
    _check_archivist_retrieval(populated_archivist_dataset)

    # the following is either not possible or not identical between archivist
    # and the old datalad-archives special remotes
    ads, akeys, archive_root, dscontent = populated_archivist_dataset

    # step 5: remove the only remaining source of the archives, and check
    # how it get/fsck fails
    for archive in archive_root.parent.glob('*.*z*'):
        archive.unlink()
    with pytest.raises(CommandError) as e:
        ads.repo.call_annex(['get', 'atar', 'azip'])
    # make sure the "reason" is communicated outwards
    assert 'does not exist' in e.value.stderr
    with pytest.raises(CommandError):
        ads.repo.call_annex(['fsck', '-f', 'archivist', 'atar', 'azip'])


def test_archivist_retrieval_legacy(
        populated_archivist_dataset, monkeypatch, no_result_rendering):
    """Same as test_archivist_retrieval(), but performs everything via the
    datalad-core provided datalad-archives special remote code
    """
    with monkeypatch.context() as m:
        m.setenv("DATALAD_ARCHIVIST_LEGACY__MODE", "yes")
        _check_archivist_retrieval(populated_archivist_dataset)


def test_claimcheck_url():
    class DummyAnnex:
        def debug(*args, **kwargs):
            pass

        def info(*args, **kwargs):
            pass

        def error(*args, **kwargs):
            pass

    ar = ArchivistRemote(DummyAnnex())

    valid_url = \
        'dl+archive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.zip#path=f.txt'
    invalid_url = \
        'dl+BROKENarchive:MD5E-s1--e9f624eb778e6f945771c543b6e9c7b2.zip#path=f.txt'

    assert ar.claimurl(valid_url) is True
    assert ar.claimurl(invalid_url) is False

    assert ar.checkurl(valid_url) is True
    assert ar.checkurl(invalid_url) is False


def test_archivist_unsupported():
    ar = ArchivistRemote(None)

    with pytest.raises(UnsupportedRequest):
        ar.transfer_store('mykey', 'myfile')
    with pytest.raises(UnsupportedRequest):
        ar.remove('mykey')



from pathlib import (
    PurePath,
)

from datalad import cfg as dlcfg

from datalad_next.datasets import Dataset
from datalad_next.utils import check_symlink_capability

from ..gitworktree import (
    GitTreeItemType,
    iter_gitworktree,
)
from ..annexworktree import iter_annexworktree

from .test_itergitworktree import prep_fp_tester


def _mkds(tmp_path_factory, monkeypatch, cfg_overrides):
    with monkeypatch.context() as m:
        for k, v in cfg_overrides.items():
            m.setitem(dlcfg.overrides, k, v)
        dlcfg.reload()
        ds = Dataset(tmp_path_factory.mktemp('ds')).create(
            result_renderer='disabled')
    dlcfg.reload()
    return ds


def _dotests(ds):
    test_file_content = 'test_file'
    test_file = ds.pathobj / 'annexed' / 'subdir' / 'file1.txt'
    test_file.parent.mkdir(parents=True)
    test_file.write_text(test_file_content)
    # we create an additional file where the content will be dropped
    # to test behavior on unavailable annex key
    droptest_content = 'somethingdropped'
    droptest_file = ds.pathobj / 'annexed' / 'dropped.txt'
    droptest_file.write_text(droptest_content)
    ds.save(result_renderer='disabled')
    ds.drop(droptest_file, reckless='availability',
            result_renderer='disabled')

    # get results for the annexed files
    query_path = ds.pathobj / 'annexed'
    res = list(iter_annexworktree(
        query_path, untracked=None, link_target=True,
    ))
    assert len(res) == 2
    #
    # pick the present annex file to start
    r = [r for r in res if r.name.name == 'file1.txt'][0]
    assert r.name == query_path / 'subdir' / 'file1.txt'
    # we cannot check gitsha and symlink content for identity, it will change
    # depending on the tuning
    # we cannot check the item type, because it will vary across repository
    # modes (e.g., adjusted unlocked)
    assert r.annexsize == len(test_file_content)
    assert r.annexkey == 'MD5E-s9--37b87ee8c563af911dcc0f949826b1c9.txt'
    # with `link_target=True` we get an objpath that is relative to the
    # query path, and we find the actual key file there
    assert (query_path / r.annexobjpath).read_text() == test_file_content
    #
    # now pick the dropped annex file
    r = [r for r in res if r.name.name == 'dropped.txt'][0]
    assert r.name == query_path / 'dropped.txt'
    # we get basic info regardless of availability
    assert r.annexsize == len(droptest_content)
    assert r.annexkey == 'MD5E-s16--770a06889bc88f8743d1ed9a1977ff7b.txt'
    # even with an absent key file, we get its would-be location,
    # and it is relative to the query path
    assert r.annexobjpath.parts[:2] == ('..', '.git')


def test_iter_annexworktree(tmp_path_factory, monkeypatch):
    ds = _mkds(tmp_path_factory, monkeypatch, {})
    _dotests(ds)


def test_iter_annexworktree_tuned(tmp_path_factory, monkeypatch):
    # same as test_file_content(), but with a "tuned" annexed that
    # no longer matches the traditional setup.
    # we need to be able to cope with that too
    ds = _mkds(tmp_path_factory, monkeypatch, {
        'annex.tune.objecthash1': 'true',
        'annex.tune.branchhash1': 'true',
        'annex.tune.objecthashlower': 'true',
    })
    _dotests(ds)


def test_iter_annexworktree_basic_fp(existing_dataset, no_result_rendering):
    ds = existing_dataset
    fcount, content_tmpl = prep_fp_tester(ds)

    for ai in filter(
        lambda i: str(i.name.name).startswith('file_'),
        iter_annexworktree(ds.pathobj, fp=True)
    ):
        fcount -= 1
        if getattr(ai, 'fp', False):
            assert content_tmpl.format(
                ai.name.name[5:]) == ai.fp.read().decode()
        else:
            assert (ai.annexobjpath and (
                ds.pathobj / ai.annexobjpath).exists() is False) or (
                    (ds.pathobj / ai.name).exists() is False)
    assert not fcount


def test_iter_annexworktree_nonrecursive(existing_dataset):
    # just a smoke test
    # given that iter_annexworktree() only wraps iter_gitworktree()
    # there is nothing to test here, any item not yielded by
    # iter_gitworktree() will also not be amended
    all_items = list(iter_annexworktree(
        existing_dataset.pathobj, recursive='no'))
    # we get a .datalad directory-tyoe item, rather than the file item from
    # inside the dir
    dirs = [i for i in all_items if i.gittype == GitTreeItemType.directory]
    assert len(dirs) == 1
    dirs[0].name == PurePath('.datalad')


def test_iter_annexworktree_noannex(existing_noannex_dataset):
    # plain smoke test to ensure this can run on a dataset without an annex
    all_annex_items = list(
        iter_annexworktree(existing_noannex_dataset.pathobj))
    all_git_items = list(iter_gitworktree(existing_noannex_dataset.pathobj))
    assert len(all_annex_items) == len(all_git_items)
    for a, g in zip(all_annex_items, all_git_items):
        assert a.name == g.name



import pytest
from .. import (
    UrlOperationsResourceUnknown,
)
from ..any import (
    _url_handlers,
    AnyUrlOperations,
)
from ..http import HttpUrlOperations
from ..file import FileUrlOperations


def test_get_best_url_handler(monkeypatch):
    ops = AnyUrlOperations()
    assert type(ops._get_handler('https://example.com')) == HttpUrlOperations
    # it will report the "longest-matching" Handler
    # we create a non-sensicle FileUrlOperations record to test that
    with monkeypatch.context() as m:
        m.setitem(
            _url_handlers,
            r'https://ex.*\.co',
            ('datalad_next.url_operations.file.FileUrlOperations',),
        )
        # the handlers are sucked into the class, so we need a new instance
        ops = AnyUrlOperations()
        assert type(ops._get_handler('https://example.com')) == FileUrlOperations


def test_any_url_operations(tmp_path):
    test_path = tmp_path / 'myfile'
    test_url = test_path.as_uri()
    ops = AnyUrlOperations()
    # no target file (yet), precise exception
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.stat(test_url)
    # now put something at the target location
    test_path.write_text('surprise!')
    # and now it works
    props = ops.stat(test_url)
    # we get the correct file size reported
    assert props['content-length'] == test_path.stat().st_size

    # and download
    download_path = tmp_path / 'download'
    props = ops.download(test_url, download_path, hash=['sha256'])
    assert props['sha256'] == '71de4622cf536ed4aa9b65fc3701f4fc5a198ace2fa0bda234fd71924267f696'
    assert props['content-length'] == 9 == test_path.stat().st_size

    # remove source and try again
    test_path.unlink()
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.download(test_url, download_path)

    # try some obscure URL scheme
    with pytest.raises(ValueError):
        ops.stat('weird://stuff')

    # and it could have been figured out before
    assert ops.is_supported_url('weird://stuff') == False



import pytest
import zipfile
from pathlib import PurePosixPath

from ..zipfile import (
    FileSystemItemType,
    ZipfileItem,
    iter_zip,
)

from ..utils import compute_multihash_from_fp


@pytest.fixture(scope="session")
def sample_zip(tmp_path_factory):
    """Create a sample zip file

    Provides a path to a zip with files and directories. Any file content is
    'zip-123\n'. The associated hashes are:

    md5: d700214df5487801e8ee23d31e60382a
    sha1: b5dfcec4d1b6166067226fae102f7fbcf6bd1bd4

    Layout::

        test-archive/
        ├── onetwothree.txt
        └── subdir/
            └── onetwothree<>again.txt
    """
    path = tmp_path_factory.mktemp('zipfile') / 'sample.zip'
    file_content = b'zip-123\n'
    with zipfile.ZipFile(path, mode='w') as zip_file:
        zip_file.writestr('test-archive/', '')
        zip_file.writestr('test-archive/subdir/', '')
        with zip_file.open('test-archive/onetwothree.txt', mode='w') as fp:
            fp.write(file_content)
        with zip_file.open('test-archive/subdir/onetwothree<>again.txt', mode='w') as fp:
            fp.write(file_content)

    yield path
    path.unlink()


def test_iter_zip(sample_zip):
    target_hash = {
        'SHA1': 'b5dfcec4d1b6166067226fae102f7fbcf6bd1bd4',
        'md5': 'd700214df5487801e8ee23d31e60382a',
    }
    root = 'test-archive'
    targets = [
        ZipfileItem(
            name=f'{root}/',
            type=FileSystemItemType.directory,
            size=0,
        ),
        ZipfileItem(
            name=f'{root}/onetwothree.txt',
            type=FileSystemItemType.file,
            size=8,
        ),
        ZipfileItem(
            name=f'{root}/subdir/',
            type=FileSystemItemType.directory,
            size=0,
        ),
        ZipfileItem(
            name=f'{root}/subdir/onetwothree<>again.txt',
            type=FileSystemItemType.file,
            size=8,
        ),
    ]

    ires = []
    for i in iter_zip(sample_zip, fp=True):
        # check that file pointer is usable
        if i.fp:
            assert compute_multihash_from_fp(
                i.fp, ['md5', 'SHA1']) == target_hash
            # we null the file pointers to ease the comparison
            i.fp = None
        ires.append(i)

    # root + subdir, 2 files
    assert 4 == len(ires)
    for r in ires:
        # do not compare mtime
        r.mtime = None
    for t in targets:
        assert t in ires



"""Tooling for test implementations

.. currentmodule:: datalad_next.tests
.. autosummary::
   :toctree: generated

   BasicGitTestRepo
   DEFAULT_BRANCH
   DEFAULT_REMOTE
   assert_in
   assert_in_results
   assert_result_count
   assert_status
   create_tree
   eq_
   get_deeply_nested_structure
   ok_
   ok_good_symlink
   ok_broken_symlink
   run_main
   skip_if_on_windows
   skip_if_root
   skip_wo_symlink_capability
   swallow_logs
   skipif_no_network
"""
# TODO `assert_raises` is excluded above to avoid syntax errors in the docstring
# rather than fixing those, we should replace it with `pytest.raises` entirely

from .utils import (
    BasicGitTestRepo,
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    assert_in,
    assert_in_results,
    assert_raises,
    assert_result_count,
    assert_status,
    create_tree,
    eq_,
    get_deeply_nested_structure,
    ok_,
    ok_good_symlink,
    ok_broken_symlink,
    run_main,
    skip_if_on_windows,
    skip_if_root,
    skip_wo_symlink_capability,
    swallow_logs,
)


from .marker import (
    skipif_no_network,
)



from contextlib import contextmanager
from pathlib import Path
from os import sep

import pytest
from datalad_next.tests import (
    BasicGitTestRepo,
    assert_raises,
    create_tree,
    get_deeply_nested_structure,
    skip_wo_symlink_capability,
    skip_if_on_windows,
    skip_if_root,
    ok_good_symlink,
    ok_broken_symlink,
    run_main,
)
from datalad_next.utils import chpwd
from datalad_next.uis import ui_switcher as ui

from datalad_next.datasets import Dataset

from ..tree import (
    Tree,
    TreeCommand
)

"""Tests for the ``datalad tree`` command."""


# ============================ Helper functions ===============================

@contextmanager
def ensure_no_permissions(path: Path):
    """Remove all permissions for given file/directory and restore the
    original permissions at the end"""

    # modeled after 'datalad.utils.ensure_write_permission'
    original_mode = path.stat().st_mode
    try:
        path.chmod(0o000)
        yield
    finally:
        try:
            path.chmod(original_mode)
        except FileNotFoundError:
            # ignore error if path was deleted in the context block
            pass


@pytest.fixture(scope="module")
def path_no_ds(tmp_path_factory):
    """Fixture for creating a temporary directory tree (**without** datasets)
    to be used in tests.

    Returns
    -------
    Path
        Root directory of the newly created tree
    """
    dir_tree = {
        "root": {
            ".dir3": {
                "dir3_file0": '',
                ".dir3_file1": '',
            },
            "dir0": {},
            "dir1": {
                "dir1_file0": '',
            },
            "dir2": {
                "dir2_dir0": {},
                "dir2_dir1": {
                    "dir2_dir1_file0": '',
                },
                "dir2_dir2": {
                    "dir2_dir2_file0": '',
                    "dir2_dir2_file1": '',
                },
                "dir2_file0": '',
                "dir2_file1": '',
            },
            ".file2": '',
            "file0": '',
            "file1": '',
        }
    }

    temp_dir_root = tmp_path_factory.mktemp("no-ds")
    create_tree(temp_dir_root, dir_tree)
    yield temp_dir_root


@pytest.fixture(scope="module")
def path_ds(tmp_path_factory):
    """Fixture for creating a temporary directory tree (**including** datasets)
    to be used in tests.

    Returns
    -------
    Path
        Root directory of the newly created tree
    """
    ds_tree = {
        "root": {
            "superds0": {
                "sd0_file0": "",
                "sd0_subds0": {
                    "sd0_sub0_subds0": {}
                }
            },
            "superds1": {
                "sd1_file0": "",
                "sd1_dir0": {
                    "sd1_d0_repo0": {},
                    "sd1_d0_subds0": {},
                },
                "sd1_ds0": {},  # not registered as subdataset
                "sd1_subds0": {},  # not installed (drop all)
            },
            # plain git repo (contents are defined in BasicGitTestRepo)
            "repo0": {},
            "file0": "",
        }
    }

    temp_dir_root = tmp_path_factory.mktemp('ds')
    create_tree(
        temp_dir_root,
        ds_tree,
    )

    # create datasets / repos
    root = temp_dir_root / "root"
    BasicGitTestRepo(path=root / "repo0", puke_if_exists=False)
    ckwa = dict(force=True, result_renderer="disabled")
    superds0 = Dataset(root / "superds0").create(**ckwa)
    sd0_subds0 = superds0.create("sd0_subds0", **ckwa)
    sd0_subds0.create("sd0_sub0_subds0", **ckwa)
    superds1 = Dataset(root / "superds1").create(**ckwa)
    superds1.create(Path("sd1_dir0") / "sd1_d0_subds0", **ckwa)
    Dataset(root / "superds1" / "sd1_ds0").create(**ckwa)
    BasicGitTestRepo(
        path=root / "superds1" / "sd1_dir0" / "sd1_d0_repo0",
        puke_if_exists=False)
    sd1_subds0 = superds1.create("sd1_subds0", **ckwa)
    sd1_subds0.drop(what='all', reckless='kill',
                    recursive=True, result_renderer='disabled')

    yield temp_dir_root


def get_tree_rendered_output(tree_cmd: list, exit_code: int = 0):
    """
    Run 'tree' CLI command with the given list of arguments and
    return the output of the custom results renderer, broken down into
    3 components (tree root, tree body, report line).

    Assumes command exit code 0 and no additional logging to stdout.

    Parameters
    ----------
    tree_cmd: list(str)
        'tree' command given as list of strings
    exit_code: int
        Expected exit code of command (default: 0)

    Returns
    -------
    Tuple[str, str, str]
        3-value tuple consisting of: tree root, tree body, report line
    """
    # remove any empty strings from command
    out, _ = run_main([c for c in tree_cmd if c != ''], exit_code=exit_code)

    # remove trailing newline
    lines = out.rstrip("\n").split("\n")

    root = lines[0]  # first line of tree output
    body = "\n".join(lines[1:-1])
    report = lines[-1]

    return root, body, report


@pytest.fixture(scope="class")
def inject_path(request, path_ds, path_no_ds):
    """
    Set a path fixture (root path of temp directory tree) as class attribute,
    to make it available to all tests in the class. The fixture is chosen based
    on the class' ``tree_with_ds`` attribute.
    """
    if request.cls.tree_with_ds:
        request.cls.path = path_ds
    else:
        request.cls.path = path_no_ds


def format_param_ids(val) -> str:
    """
    Helper to format pytest parameter IDs.

    If the parameter is a multiline string, we assume it is the
    parameter 'expected' (expected output of tree), and just
    give it a fixed ID (otherwise, it would be displayed in the
    parameter list as a long unreadable string).

    Parameters
    ----------
    val
        Parameter value
    """
    if isinstance(val, str) and "\n" in val:
        return "expected"


def build_param_matrix(matrix, params):
    """Turn inner dicts into lists (required by pytest parametrize)"""
    matrix_out = []
    for combination in matrix:
        matrix_out.append(
            # order of combinations does not matter
            [val for key, val in combination.items() if key in params]
        )
    return matrix_out


def pytest_generate_tests(metafunc):
    """Pytest helper to automatically configure parametrization.

    Avoids having to duplicate definition of parameter names and values
    across tests that use the same data.

    See: https://docs.pytest.org/en/7.1.x/example/parametrize.html#parametrizing-test-methods-through-per-class-configuration
    """
    if metafunc.cls and \
            hasattr(metafunc.cls, 'params') and \
            hasattr(metafunc.cls, 'MATRIX'):
        test_id = metafunc.function.__name__
        test_params_dict = metafunc.cls.params
        matrix = metafunc.cls.MATRIX
        if test_id in metafunc.cls.params:
            param_names = test_params_dict[test_id]
            metafunc.parametrize(
                param_names,
                build_param_matrix(matrix, param_names),
                ids=format_param_ids
            )

# ================================= Tests =====================================


@pytest.mark.usefixtures("inject_path")
class TestTree:
    """Base class with tests that should run for multiple Tree
    configurations.

    Configurations are defined by:

    - ``MATRIX``: dicts of pytest parameters and their values, where each dict
      corresponds to a separate parametrized test instance.
    - ``params``: a dict defining for each test method, which parameters
      will be used in that test (from the parameter names contained in
      ``MATRIX``).
    """
    __test__ = False  # tells pytest to not collect tests in this class
    tree_with_ds = False
    path = None  # will be set by the inject_* fixture to temp dir tree root

    # matrix of combinations of parameters to be tested and their
    # expected results
    MATRIX = []

    # dict specifying parameter sets for each test method
    params = {
        "test_print_tree": [
            "depth", "include_files", "include_hidden", "expected_str"
        ],
        "test_print_stats": [
            "depth", "include_files", "include_hidden", "expected_stats_str"
        ],
        "test_exhausted_levels_are_below_current_depth": [
            "depth", "include_files", "include_hidden"
        ]
    }


class TestTreeWithoutDatasets(TestTree):
    """Test directory tree without any datasets"""

    __test__ = True
    tree_with_ds = False

    MATRIX = [
    {
        "depth": 1,
        "include_files": False,
        "include_hidden": False,
        "expected_stats_str": "0 datasets, 3 directories",
        "expected_str": """
├── dir0/
├── dir1/
└── dir2/
"""
    },
    {
        "depth": 3,
        "include_files": False,
        "include_hidden": False,
        "expected_stats_str": "0 datasets, 6 directories",
        "expected_str": """
├── dir0/
├── dir1/
└── dir2/
    ├── dir2_dir0/
    ├── dir2_dir1/
    └── dir2_dir2/
"""
    },
    {
        "depth": 1,
        "include_files": True,
        "include_hidden": False,
        "expected_stats_str": "0 datasets, 3 directories, 2 files",
        "expected_str": """
├── dir0/
├── dir1/
├── dir2/
├── file0
└── file1
"""
    },
    {
        "depth": 3,
        "include_files": True,
        "include_hidden": False,
        "expected_stats_str": "0 datasets, 6 directories, 8 files",
        "expected_str": """
├── dir0/
├── dir1/
│   └── dir1_file0
├── dir2/
│   ├── dir2_dir0/
│   ├── dir2_dir1/
│   │   └── dir2_dir1_file0
│   ├── dir2_dir2/
│   │   ├── dir2_dir2_file0
│   │   └── dir2_dir2_file1
│   ├── dir2_file0
│   └── dir2_file1
├── file0
└── file1
"""
    },
    {
        "depth": 1,
        "include_files": True,
        "include_hidden": True,
        "expected_stats_str": "0 datasets, 4 directories, 3 files",
        "expected_str": """
├── .dir3/
├── .file2
├── dir0/
├── dir1/
├── dir2/
├── file0
└── file1
"""
    },
    {
        "depth": 3,
        "include_files": True,
        "include_hidden": True,
        "expected_stats_str": "0 datasets, 7 directories, 11 files",
        "expected_str": """
├── .dir3/
│   ├── .dir3_file1
│   └── dir3_file0
├── .file2
├── dir0/
├── dir1/
│   └── dir1_file0
├── dir2/
│   ├── dir2_dir0/
│   ├── dir2_dir1/
│   │   └── dir2_dir1_file0
│   ├── dir2_dir2/
│   │   ├── dir2_dir2_file0
│   │   └── dir2_dir2_file1
│   ├── dir2_file0
│   └── dir2_file1
├── file0
└── file1
"""
    },
    {
        "depth": 1,
        "include_files": False,
        "include_hidden": True,
        "expected_stats_str": "0 datasets, 4 directories",
        "expected_str": """
├── .dir3/
├── dir0/
├── dir1/
└── dir2/
"""
    },
    {
        "depth": 3,
        "include_files": False,
        "include_hidden": True,
        "expected_stats_str": "0 datasets, 7 directories",
        "expected_str": """
├── .dir3/
├── dir0/
├── dir1/
└── dir2/
    ├── dir2_dir0/
    ├── dir2_dir1/
    └── dir2_dir2/
"""
    },
    ]

    def test_print_tree(
            self, depth, include_files, include_hidden, expected_str
    ):
        root = str(self.path / "root")
        command = [
            'tree',
            root,
            '--depth', str(depth),
            '--include-hidden' if include_hidden else '',
            '--include-files' if include_files else ''
        ]
        _, actual_res, _ = get_tree_rendered_output(command)
        expected_res = expected_str.lstrip("\n")  # strip first newline
        ui.message("expected:")
        ui.message(expected_res)
        ui.message("actual:")
        ui.message(actual_res)
        assert expected_res == actual_res

    def test_print_stats(
            self, depth, include_files, include_hidden, expected_stats_str
    ):
        root = str(self.path / 'root')
        command = [
            'tree',
            root,
            '--depth', str(depth),
            '--include-hidden' if include_hidden else '',
            '--include-files' if include_files else ''
        ]
        _, _, actual_res = get_tree_rendered_output(command)
        expected_res = expected_stats_str
        assert expected_res == actual_res

    @pytest.mark.parametrize(
        "root_dir_name", ["root/", "root/.", "root/./", "root/../root"]
    )
    def test_root_path_is_normalized(self, root_dir_name):
        """
        Test that root path in the first line of string output
        is normalized path
        """
        root = str(self.path / root_dir_name)
        command = ['tree', root, '--depth', '0']
        actual, _, _ = get_tree_rendered_output(command)
        expected = str(self.path / "root")
        assert expected == actual

    def test_no_difference_if_root_path_absolute_or_relative(self):
        """Tree output should be identical whether the root directory
        is given as absolute or relative path"""
        root = str(self.path / "root")
        output_abs_path = get_tree_rendered_output(['tree', root])
        with chpwd(root):
            output_rel_path = get_tree_rendered_output(['tree', '.'])

        assert output_abs_path == output_rel_path

    def test_print_tree_depth_zero(self):
        root = str(self.path / "root")
        # including files should have no effect
        command = ['tree', root, '--depth', '0', '--include-files']
        actual = get_tree_rendered_output(command)
        expected = (root, '', '0 datasets, 0 directories, 0 files')
        assert expected == actual

    def test_exhausted_levels_are_below_current_depth(
            self, depth, include_files, include_hidden):
        """For each node, the exhausted levels reported for that node
        should be smaller or equal to the node's depth"""

        results = TreeCommand.__call__(
            self.path,
            depth=depth,
            include_files=include_files,
            include_hidden=include_hidden,
            result_renderer="disabled",
            # return only 'depth' and 'exhausted_levels' from result dicts
            result_xfm=lambda res: {k: res[k]
                                    for k in ("depth", "exhausted_levels")}
        )
        # sanity checks
        assert len(results) > 1
        assert any(res["exhausted_levels"] for res in results)

        # actual test
        assert all(level <= res["depth"]
                   for res in results
                   for level in res["exhausted_levels"])


class TestTreeWithDatasets(TestTreeWithoutDatasets):
    """Test directory tree with datasets"""

    __test__ = True
    tree_with_ds = True
    # set `include_files` and `include_hidden` to False,
    # they should be already covered in `TestTreeWithoutDatasets`
    MATRIX = [
    {
        "depth": 1,
        "include_files": False,
        "include_hidden": False,
        "expected_stats_str": "2 datasets, 1 directory",
        "expected_str": """
├── repo0/
├── [DS~0] superds0/
└── [DS~0] superds1/
""",
    },
    {
        "depth": 4,
        "include_files": False,
        "include_hidden": False,
        "expected_stats_str": "7 datasets, 3 directories",
        "expected_str": """
├── repo0/
├── [DS~0] superds0/
│   └── [DS~1] sd0_subds0/
│       └── [DS~2] sd0_sub0_subds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   ├── sd1_d0_repo0/
    │   └── [DS~1] sd1_d0_subds0/
    ├── [DS~0] sd1_ds0/
    └── [DS~1] (not installed) sd1_subds0/
""",
    },
    ]


class TestDatasetTree(TestTree):
    """Test dataset tree with max_dataset_depth parameter"""

    __test__ = True
    tree_with_ds = True
    MATRIX = [
    {
        "dataset_depth": 0,
        "depth": 0,
        "expected_stats_str": "3 datasets, 0 directories",
        "expected_str": """
├── [DS~0] superds0/
└── [DS~0] superds1/
    └── [DS~0] sd1_ds0/
"""
    },
    {
        "dataset_depth": 0,
        "depth": 1,
        "expected_stats_str": "3 datasets, 1 directory",
        "expected_str": """
├── [DS~0] superds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    └── [DS~0] sd1_ds0/
"""
    },
    {
        "dataset_depth": 0,
        "depth": 2,
        "expected_stats_str": "3 datasets, 2 directories",
        "expected_str": """
├── [DS~0] superds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   └── sd1_d0_repo0/
    └── [DS~0] sd1_ds0/
"""
    },
    {
        "dataset_depth": 1,
        "depth": 0,
        "expected_stats_str": "6 datasets, 1 directory",
        "expected_str": """
├── [DS~0] superds0/
│   └── [DS~1] sd0_subds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   └── [DS~1] sd1_d0_subds0/
    ├── [DS~0] sd1_ds0/
    └── [DS~1] (not installed) sd1_subds0/
"""
    },
    {
        "dataset_depth": 1,
        "depth": 2,
        "expected_stats_str": "6 datasets, 2 directories",
        "expected_str": """
├── [DS~0] superds0/
│   └── [DS~1] sd0_subds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   ├── sd1_d0_repo0/
    │   └── [DS~1] sd1_d0_subds0/
    ├── [DS~0] sd1_ds0/
    └── [DS~1] (not installed) sd1_subds0/
"""
    },
    {
        "dataset_depth": None,
        "depth": 0,
        "expected_stats_str": "7 datasets, 1 directory",
        "expected_str": """
├── [DS~0] superds0/
│   └── [DS~1] sd0_subds0/
│       └── [DS~2] sd0_sub0_subds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   └── [DS~1] sd1_d0_subds0/
    ├── [DS~0] sd1_ds0/
    └── [DS~1] (not installed) sd1_subds0/
"""
    },
    {
        "dataset_depth": None,
        "depth": 2,
        "expected_stats_str": "7 datasets, 2 directories",
        "expected_str": """
├── [DS~0] superds0/
│   └── [DS~1] sd0_subds0/
│       └── [DS~2] sd0_sub0_subds0/
└── [DS~0] superds1/
    ├── sd1_dir0/
    │   ├── sd1_d0_repo0/
    │   └── [DS~1] sd1_d0_subds0/
    ├── [DS~0] sd1_ds0/
    └── [DS~1] (not installed) sd1_subds0/
"""
    },
    ]

    params = {
        "test_print_tree": [
            "dataset_depth", "depth", "expected_str"
        ],
        "test_print_stats": [
            "dataset_depth", "depth", "expected_stats_str"
        ]
    }

    def test_print_tree(
            self, dataset_depth, depth, expected_str
    ):
        root = str(self.path / "root")

        recursive_opts = ["--recursive"]
        if dataset_depth is not None:
            recursive_opts = ['--recursion-limit', str(dataset_depth)]

        command = [
            'tree',
            root,
            '--depth', str(depth),
            *recursive_opts
        ]
        _, actual_res, _ = get_tree_rendered_output(command)
        expected_res = expected_str.lstrip("\n")  # strip first newline
        ui.message("expected:")
        ui.message(expected_res)
        ui.message("actual:")
        ui.message(actual_res)
        assert expected_res == actual_res

    def test_print_tree_without_datasets(self):
        """If there are no datasets, should only print the root"""
        root = str(self.path / "root" / "repo0")
        command = [
            'tree',
            root,
            '--depth', '10',
            '--recursive',
            '--include-files'
        ]
        _, actual_res, _ = get_tree_rendered_output(command)
        expected_res = ""
        ui.message("expected:")
        ui.message(expected_res)
        ui.message("actual:")
        ui.message(actual_res)
        assert expected_res == actual_res

    def test_print_stats(
            self, dataset_depth, depth, expected_stats_str
    ):
        root = str(self.path / "root")

        recursive_opts = ["--recursive"]
        if dataset_depth is not None:
            recursive_opts = ['--recursion-limit', str(dataset_depth)]

        command = [
            'tree',
            root,
            '--depth', str(depth),
            *recursive_opts
        ]
        _, _, actual_res = get_tree_rendered_output(command)
        expected_res = expected_stats_str
        assert expected_res == actual_res


class TestTreeFilesystemIssues:
    """Test tree with missing permissions, broken symlinks, etc."""

    def test_print_tree_fails_for_nonexistent_directory(self, tmp_path):
        """Obtain nonexistent directory by creating a temp dir and deleting it
        (may be safest method)"""
        with assert_raises(ValueError):
            Tree(tmp_path / 'nonexistent_dir', max_depth=1)

    @skip_if_root  # see https://github.com/datalad/datalad-next/issues/525
    @skip_if_on_windows
    @skip_wo_symlink_capability
    def test_print_tree_permission_denied(self, tmp_path):
        """
        - If the tree contains a directory for which the user has no
          permissions (so it would not be possible to traverse it), a message
          should be displayed next to the affected directory path
        - The rest of the tree following the forbidden directory should
          be printed as usual
        - The command should return error exit status but not crash
        """
        (tmp_path / 'z_dir' / 'subdir').mkdir(parents=True)
        forbidden_dir = tmp_path / 'a_forbidden_dir'
        forbidden_dir.mkdir(parents=True)
        # temporarily remove all permissions (octal 000)
        # restore permissions at the end, otherwise we can't delete temp dir
        with ensure_no_permissions(forbidden_dir):
            command = ['tree', str(tmp_path), '--depth', '2']
            # expect exit code 1
            _, actual, _ = get_tree_rendered_output(command, exit_code=1)
            expected = f"""
├── {forbidden_dir.name}/ [error opening dir]
└── z_dir/
    └── subdir/
""".lstrip("\n")
            ui.message("expected:")
            ui.message(expected)
            ui.message("actual:")
            ui.message(actual)
            assert expected == actual

    @skip_wo_symlink_capability
    @pytest.mark.parametrize("include_files", (True, False))
    def test_tree_with_broken_symlinks(self, tmp_path, include_files):
        """Test that broken symlinks are reported as such"""
        # prep
        dir1 = tmp_path / 'real' / 'dir1'
        file1 = tmp_path / 'real' / 'dir1' / 'file1'
        dir1.mkdir(parents=True)
        file1.touch()
        (tmp_path / 'links').mkdir()

        # create symlinks
        # 1. broken symlink pointing to non-existent target
        link_to_nonexistent = tmp_path / 'links' / '1_link_to_nonexistent'
        link_to_nonexistent.symlink_to(tmp_path / 'nonexistent')
        ok_broken_symlink(link_to_nonexistent)
        # 2. broken symlink pointing to itself
        link_to_self = tmp_path / 'links' / '2_link_to_self'
        link_to_self.symlink_to(link_to_self)
        with assert_raises((RuntimeError, OSError)):  # OSError on Windows
            # resolution should fail because of infinite loop
            link_to_self.resolve(strict=True)

        # 3. good symlink pointing to existing directory
        link_to_dir1 = tmp_path / 'links' / '3_link_to_dir1'
        link_to_dir1.symlink_to(dir1, target_is_directory=True)
        ok_good_symlink(link_to_dir1)
        # 4. good symlink pointing to existing file
        link_to_file1 = tmp_path / 'links' / '4_link_to_file1'
        link_to_file1.symlink_to(file1)
        ok_good_symlink(link_to_file1)

        # test results dict using python API
        # implicitly also tests that command yields tree without crashing
        actual = TreeCommand.__call__(
            tmp_path,
            depth=None,  # unlimited
            include_files=include_files,
            result_renderer="disabled",
            result_xfm=lambda res: (Path(res["path"]).name,
                                    res["is_broken_symlink"]),
            result_filter=lambda res: "is_broken_symlink" in res,
            return_type="list",
            on_failure="ignore"
        )

        if include_files:
            expected = [
                # (path, is_broken_symlink)
                (link_to_nonexistent.name, True),
                (link_to_self.name, True),
                (link_to_dir1.name, False),
                (link_to_file1.name, False)
            ]
        else:
            expected = [
                (link_to_dir1.name, False)
            ]
        assert set(expected) == set(actual)

    @skip_if_root  # see https://github.com/datalad/datalad-next/issues/525
    @skip_if_on_windows
    @skip_wo_symlink_capability
    @pytest.mark.parametrize("include_files", (True, False))
    def test_tree_with_broken_symlinks_to_inaccessible_targets(
            self, tmp_path, include_files):
        """Test that symlinks to targets underneath inaccessible directories
        are reported as broken, whereas symlinks to inaccessible
        file/directories themselves are not reported as broken."""
        # prep
        root = tmp_path / "root"  # tree root
        root.mkdir(parents=True)

        # create file and directory without permissions outside of tree
        # root (permissions will be removed later ad-hoc, because need to
        # create symlinks first)
        forbidden_file = tmp_path / "forbidden_file"
        forbidden_file.touch()  # permissions will be removed later ad-hoc
        forbidden_dir = tmp_path / "forbidden_dir"
        forbidden_dir.mkdir()
        file_in_forbidden_dir = forbidden_dir / "file_in_forbidden_dir"
        file_in_forbidden_dir.touch()
        dir_in_forbidden_dir = forbidden_dir / "dir_in_forbidden_dir"
        dir_in_forbidden_dir.mkdir()

        # create symlinks
        # 1. broken symlink pointing to file under inaccessible directory
        link_to_file_in_forbidden_dir = root / "1_link_to_file_in_forbidden_dir"
        link_to_file_in_forbidden_dir.symlink_to(file_in_forbidden_dir)
        with ensure_no_permissions(forbidden_dir):
            with assert_raises(PermissionError):
                # resolution should fail because of missing permissions
                link_to_file_in_forbidden_dir.resolve(strict=True)

        # 2. broken symlink pointing to directory under inaccessible directory
        link_to_dir_in_forbidden_dir = root / "2_link_to_dir_in_forbidden_dir"
        link_to_dir_in_forbidden_dir.symlink_to(dir_in_forbidden_dir)
        with ensure_no_permissions(forbidden_dir):
            with assert_raises(PermissionError):
                # resolution should fail because of missing permissions
                link_to_dir_in_forbidden_dir.resolve(strict=True)

        # 3. good symlink pointing to existing but inaccessible directory
        link_to_forbidden_dir = root / "3_link_to_forbidden_dir"
        link_to_forbidden_dir.symlink_to(forbidden_dir, target_is_directory=True)
        with ensure_no_permissions(forbidden_dir):
            ok_good_symlink(link_to_forbidden_dir)

        # 4. good symlink pointing to existing but inaccessible file
        link_to_forbidden_file = root / "4_link_to_forbidden_file"
        link_to_forbidden_file.symlink_to(forbidden_file)
        with ensure_no_permissions(forbidden_file):
            ok_good_symlink(link_to_forbidden_file)

        # temporarily remove all permissions (octal 000)
        # restore permissions at the end, otherwise we can't delete temp dir
        with ensure_no_permissions(forbidden_dir), \
                ensure_no_permissions(forbidden_file):

            # test results dict using python API
            # implicitly also tests that command yields tree without crashing
            actual = TreeCommand.__call__(
                root,
                depth=None,
                include_files=include_files,
                result_renderer="disabled",
                result_xfm=lambda res: (Path(res["path"]).name,
                                        res["is_broken_symlink"]),
                result_filter=lambda res: "is_broken_symlink" in res,
                return_type="list",
                on_failure="ignore"
            )

        if include_files:
            expected = [
                # (path, is_broken_symlink)
                (link_to_file_in_forbidden_dir.name, True),
                (link_to_dir_in_forbidden_dir.name, True),
                (link_to_forbidden_dir.name, False),
                (link_to_forbidden_file.name, False)
            ]
        else:
            expected = [
                (link_to_forbidden_dir.name, False)
            ]
        assert set(expected) == set(actual)

    @skip_wo_symlink_capability
    def test_print_tree_with_recursive_symlinks(self, tmp_path):
        """
        TODO: break down into separate tests

        - Symlinks targets are displayed in custom renderer output
        - We do not follow symlinks that point to directories underneath
          the tree root or its parent (to prevent duplicate subtrees)
        - Symlinks pointing to datasets are not considered dataset nodes
          themselves, but regular directories (to prevent duplicate counts
          of datasets)
        """
        ds = get_deeply_nested_structure(str(tmp_path / 'superds'))

        # change current dir to create symlinks with relative path
        with chpwd(ds.path):
            # create symlink to a sibling directory of the tree
            # (should be recursed into)
            (tmp_path / 'ext_dir' / 'ext_subdir').mkdir(parents=True)
            Path('link2extdir').symlink_to(Path('..') / 'ext_dir',
                                           target_is_directory=True)

            # create symlink to grandparent of the tree root (should NOT
            # be recursed into)
            Path('link2parent').symlink_to(Path('..') / '..',
                                           target_is_directory=True)

            # create symlink to subdir of the tree root at depth > max_depth
            # (should be recursed into)
            deepdir = Path('subds_modified') / 'subdir' / 'deepdir'
            deepdir.mkdir()
            (deepdir / 'subdeepdir').mkdir()
            Path('link2deepdir').symlink_to(deepdir, target_is_directory=True)

        root = ds.path
        command = ["tree", "--depth", "2", root]
        _, actual_res, counts = get_tree_rendered_output(command)
        s = sep
        expected_res = f"""
├── directory_untracked/
│   └── link2dir/ -> ..{s}subdir
├── link2deepdir/ -> subds_modified{s}subdir{s}deepdir
│   └── subdeepdir/
├── link2dir/ -> subdir
├── link2extdir/ -> ..{s}ext_dir
│   └── ext_subdir/
├── link2parent/ -> ..{s}..
├── link2subdsdir/ -> subds_modified{s}subdir
├── link2subdsroot/ -> subds_modified
├── subdir/
└── [DS~1] subds_modified/
    ├── link2superdsdir/ -> ..{s}subdir
    ├── subdir/
    └── [DS~2] subds_lvl1_modified/
""".lstrip("\n")

        # Compare with output of 'tree' command
        # ui.message(counts)
        # import subprocess
        # subprocess.run(["tree", "-dlL", "2", root])

        ui.message("expected:")
        ui.message(expected_res)
        ui.message("actual:")
        ui.message(actual_res)
        assert expected_res == actual_res



"""Common repository operations

.. currentmodule:: datalad_next.repo_utils
.. autosummary::
   :toctree: generated

   get_worktree_head
   has_initialized_annex
"""

from .annex import (
    has_initialized_annex,
)
from .worktree import (
    get_worktree_head,
)



"""Streamline user experience

Discontinue advertizing the ``hierarchy`` layout, and better explain
limitations of the command.
"""

import datalad.distributed.create_sibling_gitlab as mod_gitlab
# provide some symbols from it for the patch below
CapturedException = mod_gitlab.CapturedException
GitLabSite = mod_gitlab.GitLabSite
known_access_labels = mod_gitlab.known_access_labels
lgr = mod_gitlab.lgr

from datalad_next.commands import build_doc

from . import apply_patch

known_layout_labels = ('collection', 'flat')

command_doc = """
    Create dataset sibling at a GitLab site

    An existing GitLab project, or a project created via the GitLab web
    interface can be configured as a sibling with the :command:`siblings`
    command. Alternatively, this command can create a GitLab project at any
    location/path a given user has appropriate permissions for. This is
    particularly helpful for recursive sibling creation for subdatasets. API
    access and authentication are implemented via python-gitlab, and all its
    features are supported. A particular GitLab site must be configured in a
    named section of a python-gitlab.cfg file (see
    https://python-gitlab.readthedocs.io/en/stable/cli.html#configuration for
    details), such as::

      [mygit]
      url = https://git.example.com
      api_version = 4
      private_token = abcdefghijklmnopqrst

    Subsequently, this site is identified by its name ('mygit' in the example
    above).

    (Recursive) sibling creation for all, or a selected subset of subdatasets
    is supported with two different project layouts (see --layout):

    "flat"
      All datasets are placed as GitLab projects in the same group. The project name
      of the top-level dataset follows the configured
      datalad.gitlab-SITENAME-project configuration. The project names of
      contained subdatasets extend the configured name with the subdatasets'
      s relative path within the root dataset, with all path separator
      characters replaced by '-'. This path separator is configurable
      (see Configuration).
    "collection"
      A new group is created for the dataset hierarchy, following the
      datalad.gitlab-SITENAME-project configuration. The root dataset is placed
      in a "project" project inside this group, and all nested subdatasets are
      represented inside the group using a "flat" layout. The root datasets
      project name is configurable (see Configuration).
      This command cannot create root-level groups! To use this layout for
      a collection located in the root of an account, create the target
      group via the GitLab web UI first.

    GitLab cannot host dataset content. However, in combination with
    other data sources (and siblings), publishing a dataset to GitLab can
    facilitate distribution and exchange, while still allowing any dataset
    consumer to obtain actual data content from alternative sources.

    *Configuration*

    Many configuration switches and options for GitLab sibling creation can
    be provided arguments to the command. However, it is also possible to
    specify a particular setup in a dataset's configuration. This is
    particularly important when managing large collections of datasets.
    Configuration options are:

    "datalad.gitlab-default-site"
        Name of the default GitLab site (see --site)
    "datalad.gitlab-SITENAME-siblingname"
        Name of the sibling configured for the local dataset that points
        to the GitLab instance SITENAME (see --name)
    "datalad.gitlab-SITENAME-layout"
        Project layout used at the GitLab instance SITENAME (see --layout)
    "datalad.gitlab-SITENAME-access"
        Access method used for the GitLab instance SITENAME (see --access)
    "datalad.gitlab-SITENAME-project"
        Project "location/path" used for a datasets at GitLab instance
        SITENAME (see --project). Configuring this is useful for deriving
        project paths for subdatasets, relative to superdataset.
        The root-level group ("location") needs to be created beforehand via
        GitLab's web interface.
    "datalad.gitlab-default-projectname"
        The collection layout publishes (sub)datasets as projects
        with a custom name. The default name "project" can be overridden with
        this configuration.
    "datalad.gitlab-default-pathseparator"
        The flat and collection layout represent subdatasets with project names
        that correspond to the path, with the regular path separator replaced
        with a "-": superdataset-subdataset. This configuration can override
        this default separator.

    This command can be configured with
    "datalad.create-sibling-ghlike.extra-remote-settings.NETLOC.KEY=VALUE" in
    order to add any local KEY = VALUE configuration to the created sibling in
    the local `.git/config` file. NETLOC is the domain of the Gitlab instance to
    apply the configuration for.
    This leads to a behavior that is equivalent to calling datalad's
    ``siblings('configure', ...)``||``siblings configure`` command with the
    respective KEY-VALUE pair after creating the sibling.
    The configuration, like any other, could be set at user- or system level, so
    users do not need to add this configuration to every sibling created with
    the service at NETLOC themselves.

    """


#
# This replacement function is taken from
# https://github.com/datalad/datalad/pull/7410
# @7c83f4ac282dc3b48be8439dbbbe0f0c2c57d467
# The actual change over the patch target is only four lines, but the spaghetti
# nature of the function does not allow for a lean patch.
def _proc_dataset(refds, ds, site, project, remotename, layout, existing,
                  access, dry_run, siteobjs, depends, description):
    # basic result setup
    res_kwargs = dict(
        action='create_sibling_gitlab',
        refds=refds.path,
        path=ds.path,
        type='dataset',
        logger=lgr,
    )
    if description:
        res_kwargs['description'] = description

    if site is None:
        # always try pulling the base config from a parent dataset
        # even if paths were given (may be overwritten later)
        basecfgsite = ds.config.get('datalad.gitlab-default-site', None)

    # let the dataset config overwrite the target site, if none
    # was given
    site = refds.config.get(
        'datalad.gitlab-default-site', basecfgsite) \
        if site is None else site
    if site is None:
        # this means the most top-level dataset has no idea about
        # gitlab, and no site was specified as an argument
        # fail rather then give an error result, as this is very
        # unlikely to be intentional
        raise ValueError(
            'No GitLab site was specified (--site) or configured '
            'in {} (datalad.gitlab.default-site)'.format(ds))
    res_kwargs['site'] = site

    # determine target remote name, unless given
    if remotename is None:
        remotename_var = 'datalad.gitlab-{}-siblingname'.format(site)
        remotename = ds.config.get(
            remotename_var,
            # use config from parent, if needed
            refds.config.get(
                remotename_var,
                # fall back on site name, if nothing else can be used
                site))
    res_kwargs['sibling'] = remotename
    # check against existing remotes
    dremotes = {
        r['name']: r
        for r in ds.siblings(
            action='query',
            # fastest possible
            get_annex_info=False,
            recursive=False,
            return_type='generator',
            result_renderer='disabled')
    }
    if remotename in dremotes and existing not in ['replace', 'reconfigure']:
        # we already know a sibling with this name
        yield dict(
            res_kwargs,
            status='error' if existing == 'error' else 'notneeded',
            message=('already has a configured sibling "%s"', remotename),
        )
        return

    if layout is None:
        # figure out the layout of projects on the site
        # use the reference dataset as default, and fall back
        # on 'collection' as the most generic method of representing
        # the filesystem in a group/subproject structure
        layout_var = 'datalad.gitlab-{}-layout'.format(site)
        layout = ds.config.get(
            layout_var, refds.config.get(
                layout_var, 'collection'))
    if layout not in known_layout_labels:
        raise ValueError(
            "Unknown site layout '{}' given or configured, "
            "known ones are: {}".format(layout, known_layout_labels))

    if access is None:
        access_var = 'datalad.gitlab-{}-access'.format(site)
        access = ds.config.get(
            access_var, refds.config.get(
                access_var, 'http'))
    if access not in known_access_labels:
        raise ValueError(
            "Unknown site access '{}' given or configured, "
            "known ones are: {}".format(access, known_access_labels))

    pathsep = ds.config.get("datalad.gitlab-default-pathseparator", None) or "-"
    project_stub = \
        ds.config.get("datalad.gitlab-default-projectname", None) or "project"
    project_var = 'datalad.gitlab-{}-project'.format(site)
    process_root = refds == ds
    if project is None:
        # look for a specific config in the dataset
        project = ds.config.get(project_var, None)

    if project and process_root and layout != 'flat':
        # the root of a collection
        project = f'{project}/{project_stub}'
    elif project is None and not process_root:
        # check if we can build one from the refds config
        ref_project = refds.config.get(project_var, None)
        if ref_project:
            # layout-specific derivation of a path from
            # the reference dataset configuration
            rproject = ds.pathobj.relative_to(refds.pathobj).as_posix()
            if layout == 'collection':
                project = '{}/{}'.format(
                    ref_project,
                    rproject.replace('/', pathsep))
            else:
                project = '{}-{}'.format(
                    ref_project,
                    rproject.replace('/', pathsep))

    if project is None:
        yield dict(
            res_kwargs,
            status='error',
            message='No project name/location specified, and no configuration '
                    'to derive one',
        )
        return

    res_kwargs['project'] = project

    if dry_run:
        # this is as far as we can get without talking to GitLab
        yield dict(
            res_kwargs,
            status='ok',
            dryrun=True,
        )
        return

    # and now talk to GitLab for real
    site_api = siteobjs[site] if site in siteobjs else GitLabSite(site)

    site_project = site_api.get_project(project)
    if site_project is None:
        try:
            site_project = site_api.create_project(project, description)
            # report success
            message = "sibling repository '%s' created at %s",\
                      remotename, site_project.get('web_url', None)
            yield dict(
                res_kwargs,
                # relay all attributes
                project_attributes=site_project,
                message=message,
                status='ok',
            )
        except Exception as e:
            ce = CapturedException(e)
            yield dict(
                res_kwargs,
                # relay all attributes
                status='error',
                message=('Failed to create GitLab project: %s', ce),
                exception=ce
            )
            return
    else:
        # there already is a project
        if existing == 'error':
            # be nice and only actually error if there is a real mismatch
            if remotename not in dremotes:
                yield dict(
                    res_kwargs,
                    project_attributes=site_project,
                    status='error',
                    message=(
                        "There is already a project at '%s' on site '%s', "
                        "but no sibling with name '%s' is configured, "
                        "maybe use --existing=reconfigure",
                        project, site, remotename,
                    )
                )
                return
            elif access in ('ssh', 'ssh+http') \
                    and dremotes[remotename].get(
                        'url', None) != site_project.get(
                            # use False as a default so that there is a
                            # mismatch, complain if both are missing
                            'ssh_url_to_repo', False):
                yield dict(
                    res_kwargs,
                    project_attributes=site_project,
                    status='error',
                    message=(
                        "There is already a project at '%s' on site '%s', "
                        "but SSH access URL '%s' does not match '%s', "
                        "maybe use --existing=reconfigure",
                        project, site,
                        dremotes[remotename].get('url', None),
                        site_project.get('ssh_url_to_repo', None)
                    )
                )
                return
            elif access == 'http' \
                    and dremotes[remotename].get(
                        'url', None) != site_project.get(
                            # use False as a default so that there is a
                            # mismatch, veen if both are missing
                            'http_url_to_repo', False):
                yield dict(
                    res_kwargs,
                    project_attributes=site_project,
                    status='error',
                    message=(
                        "There is already a project at '%s' on site '%s', "
                        "but HTTP access URL '%s' does not match '%s', "
                        "maybe use --existing=reconfigure",
                        project, site,
                        dremotes[remotename].get('url', None),
                        site_project.get('http_url_to_repo', None)
                    )
                )
                return
        yield dict(
            res_kwargs,
            project_attributes=site_project,
            status='notneeded',
            message=(
                "There is already a project at '%s' on site '%s'",
                project, site,
            )
        )

    # first make sure that annex doesn't touch this one
    # but respect any existing config
    ignore_var = 'remote.{}.annex-ignore'.format(remotename)
    if ignore_var not in ds.config:
        ds.config.add(ignore_var, 'true', scope='local')

    for res in ds.siblings(
            'configure',
            name=remotename,
            url=site_project['http_url_to_repo']
            if access in ('http', 'ssh+http')
            else site_project['ssh_url_to_repo'],
            pushurl=site_project['ssh_url_to_repo']
            if access in ('ssh', 'ssh+http')
            else None,
            recursive=False,
            publish_depends=depends,
            result_renderer='disabled',
            return_type='generator'):
        yield res


apply_patch(
    'datalad.distributed.create_sibling_gitlab',
    None,
    '_proc_dataset',
    _proc_dataset)
apply_patch(
    'datalad.distributed.create_sibling_gitlab', None, 'known_layout_labels',
    known_layout_labels,
    msg='Stop advertising discontinued "hierarchy" layout for '
    '`create_siblign_gitlab()`')
# also put in effect for the constraint, add None to address limitation that
# the default also needs to be covered for datalad-core
mod_gitlab.CreateSiblingGitlab._params_['layout'].constraints._allowed = \
    (None,) + known_layout_labels

# rebuild command docs
mod_gitlab.CreateSiblingGitlab.__call__.__doc__ = None
mod_gitlab.CreateSiblingGitlab.__doc__ = command_doc
mod_gitlab.CreateSiblingGitlab = build_doc(mod_gitlab.CreateSiblingGitlab)



"""
*uncurl* git-annex external special remote
==========================================

This implementation is a git-annex accessible interface to datalad-next's URL
operations framework. It serves two main purposes:

1. Combine git-annex's capabilities of registering and accessing file content
   via URLs with DataLad's access credential management and (additional or
   alternative) transport protocol implementations.

2. Minimize the maintenance effort for datasets (primarily) composed from
   content that is remotely accessible via URLs from systems other than
   Datalad or git-annex in the event of an infrastructure transition
   (e.g. moving to a different technical system or a different data
   organization on a storage system).

Requirements
------------

This special remote implementation requires git-annex version 8.20210127 (or
later) to be available.

Download helper
---------------

The simplest way to use this remote is to initialize it without any particular
configuration::

    $ git annex initremote uncurl type=external externaltype=uncurl encryption=none
    initremote uncurl ok
    (recording state in git...)

Once initialized, or later enabled in a clone, ``git-annex addurl`` will check
with the *uncurl* remote whether it can handle a particular URL, and will let
the remote perform the download in case of positive response. By default, the
remote will claim any URLs with a scheme that the local datalad-next
installation supports. This always includes ``file://``, ``http://``, and
``https://``, but is extensible, and a particular installation may also support
``ssh://`` (by default when openssh is installed), or other schemes.

This additional URL support is also available for other commands. Here is an
example how ``datalad addurls`` can be given any uncurl-supported URLs
(here an SSH-URL) directly, provided that the ``uncurl`` remote was initialized
for a dataset (as shown above)::

    $ echo '[{"url":"ssh://my.server.org/home/me/file", "file":"dummy"}]' \\
        | datalad addurls - '{url}' '{file}'

This makes legacy commands (e.g., ``datalad download-url``), unnecessary, and
facilitates the use of more advanced ``datalad addurls`` features (e.g.,
automatic creation of subdatasets) that are not provided by lower-level
commands like ``git annex addurl``.


Download helper with credential management support
--------------------------------------------------

With this setup, download requests now also use DataLad's credential system for
authentication. DataLad will automatically lookup matching credentials, prompt
for manual entry if none are found, and offer to store them securely for later
use after having used them successfully::

    $ git annex addurl http://httpbin.org/basic-auth/myuser/mypassword
    Credential needed for access to http://httpbin.org/basic-auth/myuser/mypassword
    user: myuser
    password:
    password (repeat):
    Enter a name to save the credential
    (for accessing http://httpbin.org/basic-auth/myuser/mypassword) securely for future
    reuse, or 'skip' to not save the credential
    name: httpbin-dummy

    addurl http://httpbin.org/basic-auth/myuser/mypassword (from uncurl) (to ...)
    ok
    (recording state in git...)

By adding files via downloads from URLs in this fashion, datasets can be built
that track information across a range of locations/services, using a possibly
heterogeneous set of access methods.

This feature is very similar to the ``datalad`` special remote implementation
included in the core DataLad package. The difference here is that alternative
implementations of downloaders are employed and the ``datalad-next`` credential
system is used instead of the "providers" mechanism from DataLad's core
package.


Transforming recorded URLs
--------------------------

The main benefit of using *uncurl* is, however, only revealed when the original
snapshot of where data used to be accessible becomes invalid, maybe because
data were moved to a different storage system, or simply a different host.

This would typically require an update of each, now broken, access URL. For
datasets with thousands or even millions of files this can be an expensive
operation. For data portal operators providing a large number of datasets it is
even more tedious.

*uncurl* enables programmatic, on-access URL rewriting. This is similar, in
spirit, to Git's ``url.<base>.insteadOf`` URL modification feature. However,
modification possibilities reach substantially beyond replacing a base URL.

This feature is based on two customizable settings: 1) a *URL template*; and
2) a *set of match expressions* that extract additional identifiers
from any recorded access URL for an annex key.

Here is an example: Let's say a file in a dataset has a recorded access URL
of::

    https://data.example.org/c542/s7612_figure1.pdf

We can let *uncurl* know that ``c542`` is actually an identifier for a
particular collection of items in this data store. Likewise ``s7612`` is an
identifier of a particular item in that collection, and ``figure1.pdf`` is the
name of a component in that collection item. The following Python regular
expression can be used to "decompose" the above URL into these semantic
components::

  (?P<site>https://[^/]+)/(?P<collection>c[^/]+)/(?P<item>s[^/]+)_(?P<component>.*)$

This expression is not the most readable, but it basically chunks the URL
into segments of ``(?P<name>...)``, so-called named groups (see a
`live demo of this expression <https://www.debuggex.com/r/Aa1yua-awXBuqZ39>`__).

This expression, and additional ones like it, can set as a configuration
parameter of an *uncurl* remote setup. Extending the configuration established
by the ``initremote`` call above::

    $ git annex enableremote uncurl \\
        'match=(?P<site>https://[^/]+)/(?P<collection>c[^/]+)/(?P<item>s[^/]+)_(?P<component>.*)$'

The last argument is quoted to prevent it from being processed by the shell.

With the match expression configured, URL rewriting can be enabled by declaring
a URL template as another configuration item. The URL template uses the `Python
Format String Syntax
<https://docs.python.org/3/library/string.html#format-string-syntax>`__. If the
new URL for the file above is now
``http://newsite.net/ex-archive/c542_s7612_figure1.pdf``, we can declare
the following URL template to have *uncurl* go to the new site::

    http://newsite.net/ex-archive/{collection}_{item}_{component}

This template references the identifiers of the named groups we defined in the
match expression. Again, the URL template can be set via ``git annex
enableremote``::

    $ git annex enableremote uncurl \\
        'url=http://newsite.net/ex-archive/{collection}_{item}_{component}'

There is no need to separate the ``enableremote`` calls. Both configuration can
be given at the same time. In fact, they can also be given to ``initremote``
immediately.

The three identifiers ``site``, ``collection``, ``item``, and ``component`` are
actually a custom addition to a standard set of identifiers that are available
for composing URLs via a template.

- ``datalad_dsid`` - the DataLad dataset ID (UUID)
- ``annex_dirhash`` - "mixed" variant of the two level hash for a particular key
  (uses POSIX directory separators, and included a trailing separator)
- ``annex_dirhash_lower`` - "lower case" variant of the two level hash for a
  particular key (uses POSIX directory separators, and included a trailing
  separator)
- ``annex_key`` - git-annex key name for a request
- ``annex_remoteuuid`` - UUID of the special remote (location) used by git-annex
- ``git_remotename`` - Name of the Git remote for the uncurl special remote

.. note::
   The URL template must "resolve" to a complete and valid URL. This cannot
   be verified at configuration time, because even the URL scheme could be a
   dynamic setting.

Uploading content
-----------------

The *uncurl* special remote can upload file content or store annex keys
via supported URL schemes whenever a URL template is defined. At minimum,
storing at ``file://`` and ``ssh://`` URLs are supported. But other URL
scheme handlers with upload support may be available at a local DataLad
installation.


Deleting content
----------------

As for uploading, deleting content is only permitted with a configured
URL template. Moreover, it also depends on the delete operation being
supported for a particular URL scheme.


Configuration overrides
-----------------------

Both match expressions and the URL template can also be configured in a dataset's
configuration (committed branch configuration, or any Git configuration scope
(local, global, system) using the following configuration item names:

- ``remote.<remotename>.uncurl-url``
- ``remote.<remotename>.uncurl-match``

where ``<remotename>`` is the name of the special remote in the dataset.

A URL template provided via configuration *overrides* one defined in the special
remote setup via ``init/enableremote``.

Match expressions defined as configuration items *extend* the set of match
expressions that may be included in the special remote setup via
``init/enableremote``. The ``remote.<remotename>.uncurl-match`` configuration
item can be set as often as necessary (which one match expression each).

Tips
----

When multiple match expressions are defined, it is recommended to use unique
names for each match-group to avoid collisions.
"""
from __future__ import annotations

from annexremote import Master
from functools import partial
from pathlib import Path
import re
from typing import (
    Callable,
    Pattern,
)

from datalad_next.exceptions import (
    CapturedException,
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)
from datalad_next.url_operations import AnyUrlOperations
from datalad_next.utils import ensure_list

from . import (
    RemoteError,
    SpecialRemote,
    super_main
)


class UncurlRemote(SpecialRemote):
    """ """
    def __init__(self, annex: Master):
        super().__init__(annex)
        self.configs.update(
            url='Python format language template composing an access URL',
            match='(whitespace-separated list of) regular expression(s) to match particular components in supported URL via named groups',
        )
        self.url_tmpl: str | None = None
        self.match: list[Pattern[str]] | None = None
        self.url_handler: AnyUrlOperations | None = None
        # cache of properties that do not vary within a session
        # or across annex keys
        self.persistent_tmpl_props: dict[str, str] = {}

    def __del__(self):
        self.close()

    def close(self) -> None:
        if self.url_handler:
            del self.url_handler
            self.url_handler = None

    def initremote(self) -> None:
        # at present there is nothing that needs to be done on init/enable.
        # the remote is designed to work without any specific setup too
        pass

    def prepare(self) -> None:
        # we need the git remote name to be able to look up config about
        # that remote
        # check the config for a URL template setting
        self.url_tmpl = self.get_remote_gitcfg('uncurl', 'url', '')
        # only if we have no local, overriding, configuration ask git-annex
        # for the committed special remote config on the URL template
        if not self.url_tmpl:
            # ask for the commit config, could still be empty
            self.url_tmpl = self.annex.getconfig('url')
            # TODO test the case of a fully absent URL template
            # that would be fine and only verbatim recorded URLs could
            # be sent to a downloader

        # unconditionally ask git-annex for a match-url setting, any local
        # config amends, and does not override
        match_spec: str = self.annex.getconfig('match')
        if match_spec:
            # TODO implement sanity checks, but running it through re.compile()
            # might just be enough
            matches = [re.compile(m) for m in match_spec.split()]
        else:
            matches = []
        # extend with additional matchers from local config
        self.match = matches + [
            re.compile(m)
            for m in ensure_list(self.get_remote_gitcfg(
                'uncurl', 'match', default=[], get_all=True))
        ]

        self.message(
            f'URL rewriting template: {self.url_tmpl!r}', type='debug')
        self.message(
            f'Active URL match expressions: {[e.pattern for e in self.match]!r}',
            type='debug')

        # let the URL handler use the repo's config
        self.url_handler = AnyUrlOperations(cfg=self.repo.cfg)

        # cache template properties
        # using function arg name syntax, we need the identifiers to be valid
        # Python symbols to work in `format()`
        self.persistent_tmpl_props.update(
            datalad_dsid=self.repo.cfg.get('datalad.dataset.id', ''),
            git_remotename=self.remotename,
            annex_remoteuuid=self.annex.getuuid(),
        )

    def claimurl(self, url: str) -> bool:
        """Needs to check if want to handle a given URL

        If match expressions are configured, matches the URL against all known
        URL expressions, and returns `True` if there is any match, or
        `False` otherwise.

        If no match expressions are configured, return `True` of the URL
        scheme is supported, or `False` otherwise.
        """
        assert self.url_handler
        if self.match:
            return self.is_recognized_url(url)
        else:
            return self.url_handler.is_supported_url(url)

    def checkurl(self, url: str) -> bool:
        """
        When running `git-annex addurl`, this is called after CLAIMURL
        indicated that we could handle a URL. It can return information
        on the URL target (e.g., size of the download, a target filename,
        or a sequence thereof with additional URLs pointing to individual
        components that would jointly make up the full download from the
        given URL. However, all of that is optional, and a simple `True`
        returned is sufficient to make git-annex call `TRANSFER RETRIEVE`.
        """
        assert self.url_handler
        url_tmpl = self.url_tmpl or ''
        # try-except, because a URL template might need something
        # that cannot be extracted from this very URL.
        # we cannot consult a key that may already point to the same
        # content as the URL, and may have other information --
        # we simply dont have it at this point
        try:
            mangled_url = self.get_mangled_url(
                url,
                url_tmpl,
                self.extract_tmpl_props(
                    tmpl=url_tmpl,
                    urls=[url],
                ),
            )
        except KeyError as e:
            self.message(
                'URL rewriting template requires unavailable component '
                f'{e}, continuing with original URL',
                type='debug',
            )
            # otherwise go ahead with the original URL. the template might
            # just be here to aid structured uploads
            mangled_url = url
        try:
            self.url_handler.stat(mangled_url)
            return True
        except UrlOperationsRemoteError as e:
            # leave a trace in the logs
            CapturedException(e)
            return False
        # we could return a URL/size/filename triple instead of a bool.
        # this would make git annex download from a URL different from the input,
        # and to the given filename.
        # it could be nice to report the (mangled) url, even if the handler reports
        # a potentially deviating URL (redirects, etc.). Keeping external
        # resolvers in the loop can be intentional, and users could provide
        # resolved URL if they consider that desirable.
        # however, going with the original URL kinda does that already, rewriting
        # is happening anyways. And not reporting triplets avoids the issue
        # of git-annex insisting to write into a dedicated directory for this
        # download.

    def transfer_retrieve(self, key: str, filename: str) -> None:
        assert self.url_handler
        self._check_retrieve(
            key,
            partial(self.url_handler.download, to_path=Path(filename)),
            ('download', 'from'),
        )

    def checkpresent(self, key: str) -> bool:
        assert self.url_handler
        return self._check_retrieve(
            key,
            self.url_handler.stat,
            ('find', 'at'),
        )

    def transfer_store(self, key: str, filename: str) -> None:
        assert self.url_handler
        self._store_delete(
            key,
            partial(self.url_handler.upload, from_path=Path(filename)),
            'cannot store',
        )

    def remove(self, key: str) -> None:
        def _delete(to_url: str) -> None:
            # we have to map parameter names to be able to use a common
            # helper with transfer_store(), because UrlOperations.upload()
            # needs to get the URL as a second argument, hence we need
            # to pass parameters as keyword-args
            assert self.url_handler
            self.url_handler.delete(url=to_url)
            return

        try:
            self._store_delete(
                key,
                _delete,
                'refuses to delete',
            )
        except UrlOperationsResourceUnknown:
            self.message(
                f'{key!r} not found at the remote, skipping', type='debug')

    #
    # helpers
    #
    def is_recognized_url(self, url: str) -> bool:
        return any(m.match(url) for m in self.match or [])

    def get_key_urls(self, key: str) -> list[str]:
        # ask git-annex for the URLs it has on record for the key.
        # this will also work within checkurl() for a temporary key
        # generated by git-annex after claimurl()
        urls = self.annex.geturls(key, prefix='')
        self.message(f'Known urls for {key!r}: {urls}', type='debug')
        if self.url_tmpl:
            # we have a rewriting template. extract all properties
            # from all known URLs and instantiate the template
            # to get the ONE desired URL
            props = self.extract_tmpl_props(
                tmpl=self.url_tmpl,
                urls=urls,
                key=key,
            )
            url = self.get_mangled_url(
                fallback_url=None,
                tmpl=self.url_tmpl,
                tmpl_props=props,
            )
            return [url] if url else []
        # we have no rewriting template, and must return all URLs we know
        # to let the caller sort it out
        return urls

    def get_mangled_url(
        self,
        fallback_url: str | None,
        tmpl: str,
        tmpl_props: dict[str, str],
    ) -> str | None:
        if not tmpl:
            # the best thing we can do without a URL template is to
            # return the URL itself
            return fallback_url
        url = tmpl.format(**tmpl_props)
        return url

    def extract_tmpl_props(
        self,
        tmpl: str,
        *,
        urls: list[str] | None = None,
        key: str | None = None,
    ) -> dict[str, str]:
        # look up all the standard
        allprops = dict(self.persistent_tmpl_props)
        if key:
            allprops['annex_key'] = key
            # if we are working on a specific key, check the template if it
            # needs more key-specific properties. The conditionals below
            # are intentionally imprecise to avoid false-negatives given the
            # flexibility of the format-string-syntax
            if 'annex_dirhash' in tmpl:
                allprops['annex_dirhash'] = self.annex.dirhash(key)
            if 'annex_dirhash_lower' in tmpl:
                allprops['annex_dirhash_lower'] = self.annex.dirhash_lower(key)
        # try all URLs against all matchers
        for url in ensure_list(urls):
            for matcher in (self.match or []):
                match = matcher.match(url)
                if not match:
                    # ignore any non-match
                    continue
                # we only support named groups in expressions so this is sufficient
                props = match.groupdict()
                if any(p in allprops and allprops[p] != props[p] for p in props):
                    self.message(
                        'Partial URL property shadowing detected. '
                        'Avoid by using unique expression match group names.',
                        type='debug'
                    )
                allprops.update(props)
        return allprops

    def _check_retrieve(
        self,
        key: str,
        handler: Callable,
        action: tuple[str, str],
    ) -> bool:
        urls = self.get_key_urls(key)
        # depending on the configuration (rewriting template or not)
        # we could have one or more URLs to try
        for url in urls:
            try:
                handler(url)
                # we succeeded, no need to try again
                return True
            except UrlOperationsResourceUnknown:
                # general system access worked, but at the key location is nothing
                # to be found
                return False
            except UrlOperationsRemoteError as e:
                # return False only if we could be sure that the remote
                # system works properly and just the key is not around
                CapturedException(e)
                self.message(
                    f'Failed to {action[0]} key {key!r} {action[1]} {url!r}',
                    type='debug')
        raise RemoteError(
            f'Failed to {action[0]} {key!r} {action[1]} any of {urls!r}')

    def _store_delete(
        self,
        key: str,
        handler: Callable,
        action: str,
    ) -> None:
        if not self.url_tmpl:
            raise RemoteError(
                f'Remote {action} content without a configured URL template')
        urls = self.get_key_urls(key)
        # we have a rewriting template, so we expect exactly one URL
        assert len(urls) == 1
        url = urls[0]
        try:
            handler(to_url=url)
        except UrlOperationsResourceUnknown:
            # pass-through, would happen when removing a non-existing key,
            # which git-annex wants to be a OK thing to happen.
            # handler in callers
            raise
        except Exception as e:
            # we need to raise RemoteError whenever we could not perform
            raise RemoteError from e


_stat2checkurl_map = {
    'content-length': 'size',
}
"""Translate property names returned by AnyUrlOperations.stat()
to those expected from checkurl()"""


def main():
    """cmdline entry point"""
    super_main(
        cls=UncurlRemote,
        remote_name='uncurl',
        description=\
        "flexible access data (in archive systems) "
        "via a variety of identification schemes",
    )



"""
-- autoclass:: ShellCommandExecutor
   :special-members: __call__


"""
from __future__ import annotations

import logging
from contextlib import contextmanager
from dataclasses import dataclass
from datasalad.iterable_subprocess.iterable_subprocess import OutputFrom
from datasalad.runners.iter_subproc import iter_subproc

from queue import Queue
from typing import (
    Generator,
    Iterable,
)

from .response_generators import (
    ShellCommandResponseGenerator,
    VariableLengthResponseGenerator,
    VariableLengthResponseGeneratorPosix,
)

from datalad_next.consts import COPY_BUFSIZE
from datalad_next.exceptions import CommandError


__all__ = [
    'shell',
    'ExecutionResult',
    'ShellCommandExecutor',
]


lgr = logging.getLogger('datalad.ext.next.shell')


@dataclass
class ExecutionResult:
    stdout: bytes
    stderr: bytes
    returncode: int | None

    def to_exception(self,
                     command: bytes | str | list[str],
                     message: str = ''
                     ):
        if self.returncode != 0:
            raise CommandError(
                cmd=command.decode()
                    if isinstance(command, bytes)
                    else str(command),
                msg=message,
                returncode=self.returncode,
                stdout=self.stdout,
                stderr=self.stderr,
            )


@contextmanager
def shell(shell_cmd: list[str],
          *,
          credential: str | None = None,
          chunk_size: int = COPY_BUFSIZE,
          zero_command_rg_class: type[VariableLengthResponseGenerator] = VariableLengthResponseGeneratorPosix,
          ) -> Generator[ShellCommandExecutor, None, None]:
    """Context manager that provides an interactive connection to a shell

    This context manager uses the provided argument ``shell_cmd`` to start a
    shell-subprocess. Usually the commands provided in ``shell_cmd`` will
    start a client for a remote shell, e.g. ``ssh``.

    :func:`shell` returns an instance of :class:`ShellCommandExecutor` in the
    ``as``-variable. This instance can be used to interact with the shell. That
    means, it can be used to execute commands in the shell, receive the data
    that the commands write to their ``stdout`` and ``stderr``, and retrieve
    the return code of the executed commands. All commands that are executed
    via the returned instance of :class:`ShellCommandExecutor` are executed in
    the same shell instance.

    Parameters
    ----------
    shell_cmd : list[str]
        The command to execute the shell. It should be a list of strings that
        is given to :func:`iter_subproc` as `args`-parameter. For example:
        ``['ssh', '-p', '2222', 'localhost']``.
    chunk_size : int, optional
        The size of the chunks that are read from the shell's ``stdout`` and
        ``stderr``. This also defines the size of stored ``stderr``-content.
    zero_command_rg_class : type[VariableLengthResponseGenerator], optional, default: 'VariableLengthResponseGeneratorPosix'
        Shell uses an instance of the specified response generator class to
        execute the *zero command* ("zero command" is the command used to skip
        the login messages of the shell). This class will also be used as the
        default response generator for all further commands executed in the
        :class:`ShellCommandExecutor`-instances that is returned by
        :func:`shell`. Currently, the following concrete subclasses of
        :class:`VariableLengthResponseGenerator` exist:

            - :class:`VariableLengthResponseGeneratorPosix`: compatible with
              POSIX-compliant shells, e.g. ``sh`` or ``bash``.

            - :class:`VariableLengthResponseGeneratorPowerShell`: compatible
              with PowerShell.

    Yields
    ------
    :class:`ShellCommandExecutor`

    Examples
    --------

    **Example 1:** a simple example that invokes a single command, prints its
    output and its return code::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     result = ssh(b'ls -l /etc/passwd')
        ...     print(result.stdout)
        ...     print(result.returncode)
        ...
        b'-rw-r--r-- 1 root root 2773 Nov 14 10:05 /etc/passwd\\n'
        0

    **Example 2:** this example invokes two commands, the second of which exits
    with a non-zero return code. The error output is retrieved from
    ``result.stderr``, which contains all ``stderr`` data that was written
    since the last command was executed::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     print(ssh(b'head -1 /etc/passwd').stdout)
        ...     result = ssh(b'ls /no-such-file')
        ...     print(result.stdout)
        ...     print(result.returncode)
        ...     print(result.stderr)
        ...
        b'root:x:0:0:root:/root:/bin/bash\\n'
        b''
        2
        b"Pseudo-terminal will not be allocated because stdin is not a terminal.\\r\\nls: cannot access '/no-such-file': No such file or directory\\n"

    **Example 3:** demonstrates how to use the
    ``check``-parameter to raise a :class:`CommandError`-exception if the
    return code of the command is
    not zero. This delegates error handling to the calling code and helps to
    keep the code clean::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     print(ssh(b'ls /no-such-file', check=True).stdout)
        ...
        Traceback (most recent call last):
          File "<stdin>", line 2, in <module>
          File "/home/cristian/Develop/datalad-next/datalad_next/shell/shell.py", line 279, in __call__
            return create_result(
          File "/home/cristian/Develop/datalad-next/datalad_next/shell/shell.py", line 349, in create_result
            result.to_exception(command, error_message)
          File "/home/cristian/Develop/datalad-next/datalad_next/shell/shell.py", line 52, in to_exception
            raise CommandError(
        datalad.runner.exception.CommandError: CommandError: 'ls /no-such-file' failed with exitcode 2 [err: 'cannot access '/no-such-file': No such file or directory']

    **Example 4:** an example for manual checking of the return code::

        >>> from datalad_next.shell import shell
        >>> def file_exists(file_name):
        ...     with shell(['ssh', 'localhost']) as ssh:
        ...         result = ssh(f'ls {file_name}')
        ...         return result.returncode == 0
        ... print(file_exists('/etc/passwd'))
        True
        >>> print(file_exists('/no-such-file'))
        False

    **Example 5:** an example for result content checking::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     result = ssh(f'grep root /etc/passwd', check=True).stdout
        ...     if len(result.splitlines()) != 1:
        ...         raise ValueError('Expected exactly one line')

    **Example 6:** how to work with generator-based results.
    For long running commands a generator-based result fetching
    can be used. To use generator-based output the command has to be executed
    with the method
    :meth:`ShellCommandExecutor.start`. This method returns a generator that
    provides command output as soon as it is available::

        >>> import time
        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     result_generator = ssh.start(b'c=0; while [ $c -lt 6 ]; do head -2 /etc/passwd; sleep 2; c=$(( $c + 1 )); done')
        ...     for result in result_generator:
        ...         print(time.time(), result)
        ...     assert result_generator.returncode == 0
        1713358098.82588 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'
        1713358100.8315682 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'
        1713358102.8402972 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'
        1713358104.8490314 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'
        1713358106.8577306 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'
        1713358108.866439 b'root:x:0:0:root:/root:/bin/bash\\nsystemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin\\n'

    (The exact output of the above example might differ, depending on the
    length of the first two entries in the ``/etc/passwd``-file.)

    **Example 7:** how to use the ``stdin``-parameter to feed data to a command
    that is executed in the persistent shell.
    The methods :meth:`ShellCommandExecutor.__call__` and
    :meth:`ShellCommandExecutor.start` allow to pass an iterable in the
    ``stdin``-argument. The content of this iterable will be sent to ``stdin``
    of the executed command::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     result = ssh(b'head -c 4', stdin=(b'ab', b'c', b'd'))
        ...     print(result.stdout)
        b'abcd'

    **Example 8:** how to work with commands that consume ``stdin`` completely.
    In the previous example, the command
    ``head -c 4`` was used to consume data from ``stdin``. This command
    terminates after
    reading exactly 4 bytes from ``stdin``. If ``cat`` was used
    instead of ``head -c 4``, the command would have
    continued to run until its ``stdin`` was closed. The ``stdin`` of the
    command that is executed in the persistent shell can be close by calling
    :meth:`ssh.close`. But, in order to be able to call :meth:`ssh.close`,
    any process that consumes ``stdin`` completely should be executed by
    calling the :meth:`ssh.start`-method.
    The reason for this is that :meth:`ssh.start` will return immediately which
    allows to call the :meth:`ssh.close`-method, as shown in the following
    code (:meth:`ssh.__call__` would have waited for ``cat`` to terminate, but
    because :meth:`ssh.close` is not called, ``cat`` would never terminate)::

        >>> from datalad_next.shell import shell
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     result_generator = ssh.start(b'cat', stdin=(b'12', b'34', b'56'))
        ...     ssh.close()
        ...     print(tuple(result_generator))
        (b'123456',)

    Note that
    the ``ssh``-object cannot be used for further command execution after
    :meth:`ssh.close` was called. Further command execution requires to spin up
    a new persistent shell-object. To prevent this overhead, it is advised to
    limit the number of bytes that a shell-command consumes, either by their
    number, e.g. by using ``head -c``, or by some other means, e.g.
    by interpreting the content or using a command like ``timeout``.

    **Example 9:** upload a file to the persistent shell. The command
    ``head -c`` can be used to implement the upload a file to a remote shell.
    The basic idea
    is to determine the number of bytes that will be uploaded and create a
    command in the remote shell that will consume exactly this amount of bytes.
    The following code implements this idea (without file-name escaping and
    error handling)::

        >>> import os
        >>> import time
        >>> from datalad_next.shell import shell
        >>> def upload(ssh, file_name, remote_file_name):
        ...     size = os.stat(file_name).st_size
        ...     f = open(file_name, 'rb')
        ...     return ssh(f'head -c {size} > {remote_file_name}', stdin=iter(f.read, b''))
        ...
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     upload(ssh, '/etc/passwd', '/tmp/uploaded-1')

    Note: in this example, ``f`` is not explicitly closed, it is only
    closed when the program exits. The reason for
    this is that the shell uses threads internally for stdin-feeding, and there
    is no simple way to determine whether the thread that reads ``f`` has yet
    read an EOF and exited. If ``f`` is closed before the thread exits, and the
    thread tries to read from ``f``, a ``ValueError`` will be raised (the
    function :func:`datalad_next.shell.posix.upload` contains a solution
    for this problem that has slightly more code. For the sake of simplicity,
    this solution was not implemented in the example above).

    **Example 10:** download a file. This example
    uses a fixed-length response generator
    to download a file from a remote shell. The basic idea is to determine the
    number of bytes that will be downloaded and create a fixed-length response
    generator that reads exactly this number of bytes. The fixed length response
    generator is then passed to :meth:`ssh.start` in the keyword-argument
    ``response_generator``. This instructs :meth:`ssh.start` to use the response
    generator to interpret the output of this command invocation (the example
    code has no file-name escaping or error handling)::

        >>> from datalad_next.shell import shell
        >>> from datalad_next.shell.response_generators import FixedLengthResponseGeneratorPosix
        >>> def download(ssh, remote_file_name, local_file_name):
        ...     size = ssh(f'stat -c %s {remote_file_name}').stdout
        ...     with open(local_file_name, 'wb') as f:
        ...         response_generator = FixedLengthResponseGeneratorPosix(ssh.stdout, int(size))
        ...         results = ssh.start(f'cat {remote_file_name}', response_generator=response_generator)
        ...         for chunk in results:
        ...             f.write(chunk)
        ...
        >>> with shell(['ssh', 'localhost']) as ssh:
        ...     download(ssh, '/etc/passwd', '/tmp/downloaded-1')
        ...

    Note that :meth:`ssh.start` is used to start the download. This allows to
    process downloaded data as soon as it is available.

    **Example 11:**
    This example implements interaction with a *Python* interpreter (which
    can be local or remote). Interaction in the context of this example means,
    executing a
    line of python code, returning the result, i.e. the output on ``stdout``,
    and detect whether an exception was raised or not. To this end
    a Python-specific variable-length response generator is created by
    subclassing the
    generic class :class:`VariableLengthResponseGenerator`. The new response
    generator implements the method :meth:`get_final_command`, which takes a
    python statement and returns a ``try``-``except``-block that executes the
    python statement, prints the end-marker and a return code (which is ``0`` if
    the statement was executed successfully, and ``1`` if an exception was
    raised)::

        >>> from datalad_next.shell import shell
        >>> from datalad_next.shell.response_generators import VariableLengthResponseGenerator
        >>> class PythonResponseGenerator(VariableLengthResponseGenerator):
        ...     def get_final_command(self, command: bytes) -> bytes:
        ...         return f'''try:
        ...     {command.decode()}
        ...     print('{self.end_marker.decode()}')
        ...     print(0)
        ... except:
        ...     print('{self.end_marker.decode()}')
        ...     print(1)
        ... '''.encode()
        ...     @property
        ...     def zero_command(self) -> bytes:
        ...         return b'True'
        ...
        >>> with shell(['python', '-u', '-i']) as py:
        ...     print(py('1 + 1'))
        ...     print(py('1 / 0'))
        ...
        ExecutionResult(stdout=b'2\\n', stderr=b'>>> ... ... ... ... ... ... ... ... ', returncode=0)
        ExecutionResult(stdout=b'', stderr=b'... ... ... ... ... ... ... ... Traceback (most recent call last):\\n  File "<stdin>", line 2, in <module>\\nZeroDivisionError: division by zero', returncode=1)

    The python response generator could be extended to deliver exception
    information in an extended ``ExecutionResult``. This can be achieved by
    *pickling* (see the ``pickle``-module) a caught exception to a byte-string,
    printing this byte-string after the return-code line, and printing another
    end-marker. The :meth:`send`-method of the response generator must then
    be overwritten to unpickle the exception information and store it in an
    extended ``ExecutionResult`` (or raise it in the shell-context, if that is
    preferred).

    **Example 12:** this example shows how to use the shell context handler
    in situations were a ``with``-statement is not suitable, e.g. if a shell
    object should be used in multiple, independently called functions. In this
    case the context manager
    can be manually entered and exited. The following code generates a global
    ``ShellCommandExecutor``-instance in the ``ssh``-variable::

        >>> from datalad_next.shell import shell
        >>> context_manager = shell(['ssh', 'localhost'])
        >>> ssh = context_manager.__enter__()
        >>> print(ssh(b'ls /etc/passwd').stdout)
        b'/etc/passwd\\n'
        >>> context_manager.__exit__(None, None, None)
        False

    """

    def train(queue: Queue):
        """Use a queue to allow chaining of iterables at different times"""
        for iterable in iter(queue.get, None):
            yield from iterable

    subprocess_inputs: Queue = Queue()
    with iter_subproc(shell_cmd,
                      inputs=train(subprocess_inputs),
                      chunk_size=chunk_size,
                      bufsize=0) as shell_output:

        assert issubclass(zero_command_rg_class, VariableLengthResponseGenerator)

        cmd_executor = ShellCommandExecutor(
            subprocess_inputs,
            shell_output,
            shell_cmd,
            zero_command_rg_class
        )
        try:
            cmd_executor.command_zero(zero_command_rg_class(shell_output))
            # Return the now ready connection
            yield cmd_executor
        finally:
            # Ensure that the shell is terminated if an exception is raised by
            # code that uses `shell`. This is necessary because
            # the `terminate`-call that is invoked when leaving the
            # `iterable_subprocess`-context will not end the shell-process. It
            # will only terminate if its stdin is closed, or if it is killed.
            subprocess_inputs.put(None)


class ShellCommandExecutor:
    """Execute a command in a shell and return a generator that yields output

    Instances of :class:`ShellCommandExecutor` allow to execute commands that
    are provided as byte-strings via its :meth:`__call__`-method.

    To execute the command and collect its output,
    return code, and stderr-output, :class:`ShellCommandExecutor` uses
    instances of subclasses of :class:`ShellCommandResponseGenerator`, e.g.
    :class:`VariableLengthResponseGeneratorPosix`.
    """

    def __init__(self,
                 process_inputs: Queue,
                 stdout: OutputFrom,
                 shell_cmd: list[str],
                 default_rg_class: type[VariableLengthResponseGenerator],
                 ) -> None:
        self.process_inputs = process_inputs
        self.stdout = stdout
        self.shell_cmd = shell_cmd
        self.default_rg_class = default_rg_class

    def __call__(self,
                 command: bytes | str,
                 *,
                 stdin: Iterable[bytes] | None = None,
                 response_generator: ShellCommandResponseGenerator | None = None,
                 encoding: str = 'utf-8',
                 check: bool = False
                 ) -> ExecutionResult:
        """Execute a command in the connected shell and return the result

        This method executes the given command in the connected shell. It
        assembles all output on stdout, all output on stderr that was
        written during the execution of the command, and the return
        code of the command.
        (The response generator defines when the command output is considered
        complete. Usually that is done by checking for a random end-of-output
        marker.)

        Parameters
        ----------
        command : bytes | str
            The command to execute. If the command is given as a string, it
            will be encoded to bytes using the encoding given in `encoding`.
        stdin : Iterable[byte] | None, optional, default: None
            If given, the bytes are sent to stdin of the command.

            Note: If the command reads its ``stdin`` until EOF, you have to use
            :meth:`self.close` to close ``stdin`` of the command. Otherwise,
            the command will usually not terminate. Once :meth:`self.close` is
            called, no more commands can be executed with this
            :class:`ShellCommandExecutor`-instance. If you want to execute
            further commands in the same :class:`ShellCommandExecutor`-instance,
            you must ensure
            that commands consume a fixed amount of input, for example,
            by using `head -c <byte-count> | <command>`.
        response_generator : ShellCommandResponseGenerator | None, optional, default: None
            If given, the responder generator (usually an instance of a subclass
            of ``ShellCommandResponseGenerator``), that is used to generate the
            command line and to parse the output of the command. This can be
            used to implement, for example, fixed length output processing.
        encoding : str, optional, default: 'utf-8'
            The encoding that is used to encode the command if it is given as a
            string. Note: the encoding should match the decoding the is used in
            the connected shell.
        check : bool, optional, default: False
            If True, a :class:`CommandError`-exception is raised if the return
            code of the command is not zero.

        Returns
        -------
        :class:`ExecutionResult`
            An instance of :class:`ExecutionResult` that contains the
            ``stdout``-output, the ``stderr``-output, and the return code of
            the command.

        Raises
        ------
        :class:`CommandError`
            If the return code of the command is not zero and `check` is True.
        """
        response_generator = self.start(
            command,
            stdin=stdin,
            response_generator=response_generator,
            encoding=encoding,
        )
        stdout = b''.join(response_generator)
        stderr = b''.join(self.stdout.stderr_deque)
        self.stdout.stderr_deque.clear()
        return create_result(
            response_generator,
            command,
            stdout,
            stderr,
            check=check
        )

    def start(self,
              command: bytes | str,
              *,
              stdin: Iterable[bytes] | None = None,
              response_generator: ShellCommandResponseGenerator | None = None,
              encoding: str = 'utf-8',
              ) -> ShellCommandResponseGenerator:
        """Execute a command in the connected shell

        Execute a command in the connected shell and return a generator that
        provides the content written to stdout of the command. After the
        generator is exhausted, the return code of the command is available
        in the ``returncode``-attribute of the generator.

        Parameters
        ----------
        command : bytes | str
            The command to execute. If the command is given as a string, it
            will be encoded to bytes using the encoding given in `encoding`.
        stdin : Iterable[byte] | None, optional, default: None
            If given, the bytes are sent to stdin of the command.

            Note: If the command reads its ``stdin`` until EOF, you have to use
            :meth:`self.close` to close ``stdin`` of the command. Otherwise,
            the command will usually not terminate. Once :meth:`self.close` is
            called, no more commands can be executed with this
            :class:`ShellCommandExecutor`-instance. If you want to execute
            further commands in the same :class:`ShellCommandExecutor`-instance,
            you must ensure
            that commands consume a fixed amount of input, for example,
            by using `head -c <byte-count> | <command>`.
        response_generator : ShellCommandResponseGenerator | None, optional, default: None
            If given, the responder generator (usually an instance of a subclass
            of ``ShellCommandResponseGenerator``), that is used to generate the
            command line and to parse the output of the command. This can be
            used to implement, for example, fixed length output processing.
        encoding : str, optional, default: 'utf-8'
            The encoding that is used to encode the command if it is given as a
            string. Note: the encoding should match the decoding the is used in
            the connected shell.

        Returns
        -------
        :class:`ShellCommandResponseGenerator`

            A generator that yields the output of ``stdout`` of the command.
            The generator is exhausted when all output is read. After that,
            the return code of the command execution
            is available in the ``returncode``-attribute of the generator,
            and the stderr-output is available in the ``stderr_deque``-attribute
            of the response generator.
            If a response generator was passed in via the
            ``response_generator``-parameter, the same instance will be
            returned.
        """
        if response_generator is None:
            response_generator = self.default_rg_class(self.stdout)

        if isinstance(command, str):
            command = command.encode(encoding)

        final_command = response_generator.get_final_command(command)
        # Store the command list to report it in `CommandError`-exceptions.
        # This is done here to relieve the response generator classes from
        # this task.
        self.process_inputs.put([final_command])
        if stdin is not None:
            self.process_inputs.put(stdin)
        return response_generator

    def __repr__(self):
        return f'{self.__class__.__name__}({self.shell_cmd!r})'

    def close(self):
        """stop input to the shell

        This method closes stdin of the shell. This will in turn terminate
        the shell, no further commands can be executed in the shell.
        """
        self.process_inputs.put(None)

    def command_zero(self,
                     response_generator: VariableLengthResponseGenerator
                     ) -> None:
        """Execute the zero command

        This method is only used by :func:`shell` to skip any login messages
        """
        result_zero = self(
            response_generator.zero_command,
            response_generator=response_generator,
            check=True,
        )
        lgr.debug('skipped login message: %s', result_zero.stdout)


def create_result(response_generator: ShellCommandResponseGenerator,
                  command: bytes | str | list[str],
                  stdout: bytes,
                  stderr: bytes,
                  error_message: str = '',
                  check: bool = False) -> ExecutionResult:

    result = ExecutionResult(
        stdout=stdout,
        stderr=stderr,
        returncode=response_generator.returncode
    )
    if check is True:
        result.to_exception(command, error_message)
    return result



from pathlib import (
    Path,
    PurePath,
)
import pytest

from datalad.api import ls_file_collection

from datalad_next.constraints import CommandParametrizationError
# we need this fixture
from datalad_next.iter_collections.tests.test_iterzip import sample_zip
from datalad_next.tests import skipif_no_network

from ..ls_file_collection import LsFileCollectionParamValidator


def test_ls_file_collection_insufficient_args():
    with pytest.raises(CommandParametrizationError):
        ls_file_collection()

    # any collection needs some kind of identifier, just the type
    # parameter is not enough
    with pytest.raises(CommandParametrizationError):
        ls_file_collection('tarfile')

    # individual collection types have particular requirements re
    # the identifiers -- tarfile wants an existing path
    with pytest.raises(CommandParametrizationError):
        ls_file_collection('tarfile', 'http://example.com')

    # not a known collection type
    with pytest.raises(CommandParametrizationError):
        ls_file_collection('bogus', 'http://example.com')


def _check_archive_member_result(r, collection):
    # basics of a result
    assert r['action'] == 'ls_file_collection'
    assert r['status'] == 'ok'
    # a collection identifier, here the tar location
    assert 'collection' in r
    assert r['collection'] == collection
    # an item identifier, here a str-path of an archive member
    assert 'item' in r
    assert isinstance(r['item'], str)
    # item type info, here some filesystem-related category
    assert 'type' in r
    assert r['type'] in ('file', 'directory', 'symlink', 'hardlink')


def test_ls_file_collection_zipfile(sample_zip, no_result_rendering):
    for res in (
        ls_file_collection('zipfile', sample_zip),
        ls_file_collection('zipfile', sample_zip, hash='md5'),
    ):
        assert len(res) == 4
        # test a few basic properties that should be true for any result
        for r in res:
            _check_archive_member_result(r, sample_zip)


@skipif_no_network
def test_ls_file_collection_tarfile(sample_tar_xz, no_result_rendering):
    for res in (
        ls_file_collection('tarfile', sample_tar_xz),
        ls_file_collection('tarfile', sample_tar_xz, hash='md5'),
    ):
        assert len(res) == 6
        # test a few basic properties that should be true for any result
        for r in res:
            _check_archive_member_result(r, sample_tar_xz)


def test_ls_file_collection_directory(tmp_path, no_result_rendering):
    # smoke test on an empty dir
    res = ls_file_collection('directory', tmp_path)
    assert len(res) == 0


def test_ls_file_collection_gitworktree(existing_dataset, no_result_rendering):
    # smoke test on a plain dataset
    res = ls_file_collection('gitworktree', existing_dataset.pathobj)
    assert len(res) > 1
    assert all('gitsha' in r for r in res)

    # and with hashing
    res_hash = ls_file_collection('gitworktree', existing_dataset.pathobj,
                                  hash='md5')
    assert len(res) == len(res_hash)
    assert all('hash-md5' in r for r in res_hash)


def test_ls_file_collection_validator():
    val = LsFileCollectionParamValidator()

    with pytest.raises(RuntimeError):
        val.get_collection_iter(type='bogus', collection='any', hash=None)


@skipif_no_network
def test_replace_add_archive_content(sample_tar_xz, existing_dataset,
                                     no_result_rendering):
    ds = existing_dataset
    archive_path = ds.pathobj / '.datalad' / 'myarchive.tar.xz'
    # get archive copy in dataset (not strictly needed, but
    # add-archive-content worked like this
    ds.download({sample_tar_xz.as_uri(): archive_path})
    # properly safe to dataset (download is ignorant of datasets)
    res = ds.save(message='add archive')
    # the first result has the archive addition, snatch the archive key from it
    assert res[0]['path'] == str(archive_path)
    archive_key = res[0]['key']

    # now we can scan the archive and register keys for its content.
    # the order and specific composition of the following steps is flexible.
    # we could simply extract the local archive, save the content to the
    # dataset, and then register `dl+archive` URLs.
    # however, we will use an approach that does not require any data
    # to be present locally (actually not even the archive that we have locally
    # already for this test), but is instead based on some metadata
    # that is provided by `ls-file-collection` (but could come from elsewhere,
    # including `ls-file-collection` executed on a different host).
    file_recs = [
        r for r in ls_file_collection(
            'tarfile', sample_tar_xz, hash=['md5'],
        )
        # ignore any non-file, would not have an annex key.
        # Also ignores hardlinks (they consume no space (size=0), but could be
        # represented as regular copies of a shared key. however, this
        # requires further processing of the metadata records, in order to find
        # the size of the item that has the same checksum as this one)
        if r.get('type') == 'file'
    ]
    # we enable the `datalad-archives` special remote using a particular
    # configuration that `add-archive-content` would use.
    # this special remote can act on the particular URLs that we will add next
    ds.repo.call_annex([
        'initremote', 'datalad-archives', 'type=external',
        'externaltype=datalad-archives', 'encryption=none', 'autoenable=true'])
    # assign special `dl+archive` URLs to all file keys
    # the `datalad-archives` special remote will see them and perform the
    # extraction of file content from the archive on demand.
    # the entire operation is not doing any extraction or data retrieval,
    # because we have all information necessary to generate keys
    ds.addurls(
        # takes an iterable of dicts
        file_recs,
        # urlformat: handcrafted archive key, as expected by datalad-archive
        # (double braces to keep item and size as placeholders for addurls)
        f'dl+archive:{archive_key}#path={{item}}&size={{size}}',
        # filenameformat
        '{item}',
        key='et:MD5-s{size}--{hash-md5}',
    )
    # because we have  been adding the above URLs using a pure metadata-driven
    # approach, git-annex does not yet know that the archives remote actually
    # has the keys. we could use `annex setpresentkey` for that (fast local
    # operation), but here we use `fsck` to achieve a comprehensive smoke test
    # of compatibility with our hand-crafted and the special remote
    # implementation
    # (actually: without --fast the special remote crashes with a protocol
    #  error -- a bug in the special remote probably)
    ds.repo.call_annex(
        ['fsck', '--fast', '-f', 'datalad-archives'],
        files=['test-archive'],
    )
    # at this point we are done
    # check retrieval for a test file, which is not yet around
    testfile = ds.pathobj / 'test-archive' / '123_hard.txt'
    assert ds.status(
        testfile, annex='availability')[0]['has_content'] is False
    ds.get(testfile)
    assert testfile.read_text() == '123\n'


def test_ls_renderer():
    # nothing more than a smoke test
    ls_file_collection(
        'directory',
        Path(__file__).parent,
        result_renderer='tailored',
    )


def test_ls_annexworktree_empty_dataset(existing_dataset):
    res = ls_file_collection(
        'annexworktree',
        existing_dataset.pathobj,
        result_renderer='disabled'
    )
    assert len(res) == 3
    annexed_files = [annex_info for annex_info in res if 'annexkey' in annex_info]
    assert len(annexed_files) == 0


def test_ls_annexworktree_simple_dataset(existing_dataset):

    (existing_dataset.pathobj / 'sample.bin').write_bytes(b'\x00' * 1024)
    existing_dataset.save(message='add sample file')

    res = ls_file_collection(
        'annexworktree',
        existing_dataset.pathobj,
        result_renderer='disabled'
    )
    assert len(res) == 4
    annexed_files = [annex_info for annex_info in res if 'annexkey' in annex_info]
    assert len(annexed_files) == 1
    assert annexed_files[0]['type'] == 'annexed file'
    assert {
        'annexkey',
        'annexsize',
        'annexobjpath'
    }.issubset(set(annexed_files[0].keys()))



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Helper for parameter validation, documentation and conversion"""

__docformat__ = 'restructuredtext'

import re


def _strip_typerepr(s):
    """Strip away <class '...'> and <type '...'> decorations for docstrings
    """
    return re.sub(r"<(class|type) '(\S+)'>", r'\2', s)


def _type_str(t):
    """Get string human-readable representation of a data type

    If type (t) is given as a tuple, assume ability to choose any of the
    listed types, so those types listing get joined with |
    """
    if isinstance(t, tuple):
        s = ' or '.join(map(_type_str, t))
        return ("(%s)" % s) if len(t) > 1 else s
    return _strip_typerepr(str(t))



import pytest

from ..  import ParamDictator


def test_paramdictator():
    d = {'a': 1, 'b': 2}
    pd = ParamDictator(d)
    assert pd.a == 1
    assert pd.b == 2
    with pytest.raises(AssertionError):
        assert pd.__dir__ is None



"""Provide a full replacement of `SSHRemoteIO`

First and foremost, this replacement no longer uses the remote shell
implementation of the previous version, but is based on `datalad_next.shell`.

Moreover, the ``cmd``-argument for the shell ssh-process, is not correct, if
``self.ssh`` is an instance of ``NoMultiplexSSHConnection``.

The changes in this patch build the correct ``cmd``-argument by adding
additional arguments to ``cmd``, if `self.ssh` is an instance of
``NoMultiplexSSHConnection``. More precisely, the arguments that are
required to open a "shell" in a ``NoMultiplexSSHConnection`` are stored in
``NoMultiplexSSHConnection._ssh_open_args`` and not in
``NoMultiplexSSHConnection._ssh_args``. This patch therefore provides
arguments from both lists, i.e. from ``_ssh_args`` and ``_ssh_open_args`` in
the call that opens a "shell", if ``self.ssh`` is an instance of
``NoMultiplexSSHConnection``.

The implementation also no longer assumes that local and remote platform are
identical. This patch introduces an actual remote platform/system
determination.

This patch also adds the method :meth:`url2transport_path`, which is used to
convert abstract paths, which are used in the patched RIA/ORA-code, into paths
that SSHRemoteIO can operate on.
"""
from __future__ import annotations
from pathlib import (
    Path,
    PurePosixPath,
)
from urllib.parse import (
    unquote,
    urlparse,
)

from datalad.distributed.ora_remote import (
    DEFAULT_BUFFER_SIZE,
    IOBase,
    RemoteError,
    RIARemoteError,
    contextmanager,
    functools,
    on_osx,
    sh_quote,
    ssh_manager,
    stat,
)
from datalad.support.sshconnector import NoMultiplexSSHConnection

from datalad_next.exceptions import CapturedException
from datalad_next.patches import apply_patch
from datalad_next.runners import CommandError
from datalad_next.shell import (
    FixedLengthResponseGeneratorPosix,
    shell,
    posix as posix_ops,
)


class SSHRemoteIO(IOBase):
    """IO operation if the object tree is SSH-accessible

    It doesn't even think about a windows server.
    """
    def __init__(self, ssh_url, buffer_size=DEFAULT_BUFFER_SIZE):
        """
        Parameters
        ----------
        ssh_url : str
          SSH-accessible host(name) to perform remote IO operations
          on.
        buffer_size: int or None
          The buffer size to be used as the `chunk_size` for communication
          with the remote shell.
        """
        parsed_url = urlparse(ssh_url)

        self.url = ssh_url
        self._remote_system = None
        # the connection to the remote
        # we don't open it yet, not yet clear if needed
        self.ssh = ssh_manager.get_connection(
            ssh_url,
            use_remote_annex_bundle=False,
        )
        self.ssh.open()

        ssh_args = self.ssh._ssh_args
        if isinstance(self.ssh, NoMultiplexSSHConnection):
            ssh_args.extend(self.ssh._ssh_open_args)
        cmd = ['ssh'] + ssh_args + [self.ssh.sshri.as_str()]

        # we settle on `bash` as a shell. It should be around and then we
        # can count on it
        cmd.append('bash')
        # open the remote shell
        self.servershell_context = shell(
            cmd,
            chunk_size=buffer_size,
        )
        self.servershell = self.servershell_context.__enter__()

        # if the URL had a path, we try to 'cd' into it to make operations on
        # relative paths intelligible
        if parsed_url.path:
            # unquote path
            real_path = unquote(parsed_url.path)
            try:
                self.servershell(
                    f'cd {sh_quote(real_path)}',
                    check=True,
                )
            except Exception as e:
                # it may be a legit use case to point to a directory that is
                # not yet existing. Log and continue
                CapturedException(e)

    def close(self):
        if self.servershell_context is None:
            return
        self.servershell_context.__exit__(None, None, None)
        self.servershell_context = None

    def url2transport_path(
            self,
            url_path: PurePosixPath
    ) -> Path | PurePosixPath:
        assert isinstance(url_path, PurePosixPath)
        return url_path

    @property
    def remote_system(self):
        if self._remote_system is None:
            self._remote_system = self.servershell(
                "uname -s",
                check=True
            ).stdout.strip().decode().casefold()
        return self._remote_system

    @contextmanager
    def ensure_writeable(self, path):
        """Context manager to get write permission on `path` and restore
        original mode afterwards.

        If git-annex ever touched the key store, the keys will be in mode 444
        directories, and we need to obtain permission first.

        Parameters
        ----------
        path: Path
          path to the target file
        """
        path = sh_quote(str(path))
        # remember original mode -- better than to prescribe a fixed mode

        if self.remote_system == 'darwin':
            format_option = "-f%Dp"
            # on macOS this would return decimal representation of mode (same
            # as python's stat().st_mode
            conversion = int
        else:  # win is currently ignored anyway
            format_option = "--format=\"%f\""
            # in opposition to the above form for macOS, on debian this would
            # yield the hexadecimal representation of the mode; hence conversion
            # needed.
            conversion = functools.partial(int, base=16)

        output = self.servershell(
            f"stat {format_option} {path}",
            check=True,
        ).stdout.decode()
        mode = conversion(output)
        if not mode & stat.S_IWRITE:
            new_mode = oct(mode | stat.S_IWRITE)[-3:]
            self.servershell(f"chmod {new_mode} {path}", check=True)
            changed = True
        else:
            changed = False
        try:
            yield
        finally:
            if changed:
                # restore original mode
                self.servershell(
                    f"chmod {oct(mode)[-3:]} {path}",
                    # don't fail if path doesn't exist anymore
                    check=False,
                )

    def mkdir(self, path):
        self.servershell(
            f'mkdir -p {sh_quote(str(path))}',
            check=True,
        )

    def symlink(self, target, link_name):
        self.servershell(
            f'ln -s {sh_quote(str(target))} {sh_quote(str(link_name))}',
            check=True,
        )

    def put(self, src, dst, progress_cb):
        posix_ops.upload(
            self.servershell,
            Path(src),
            PurePosixPath(dst),
            # the given callback only takes a single int, but posix.upload
            # gives two (cur, target) -> have an adaptor
            progress_callback=lambda c, m: progress_cb(c),
            check=True,
        )

    def get(self, src, dst, progress_cb):
        posix_ops.download(
            self.servershell,
            PurePosixPath(src),
            Path(dst),
            # the given callback only takes a single int, but posix.download
            # gives two (cur, target) -> have an adaptor
            progress_callback=lambda c, m: progress_cb(c),
            check=True,
        )

    def rename(self, src, dst):
        with self.ensure_writeable(dst.parent):
            self.servershell(
                f'mv {sh_quote(str(src))} {sh_quote(str(dst))}',
                check=True,
            )

    def remove(self, path):
        try:
            with self.ensure_writeable(path.parent):
                self.servershell(
                    f'rm {sh_quote(str(path))}',
                    check=True,
                )
        except CommandError as e:
            raise RIARemoteError(
                f"Unable to remove {path} "
                "or to obtain write permission in parent directory.") from e

    def remove_dir(self, path):
        with self.ensure_writeable(path.parent):
            self.servershell(
                f'rmdir {sh_quote(str(path))}',
                check=True,
            )

    def exists(self, path):
        try:
            self.servershell(
                f'test -e {sh_quote(str(path))}',
                check=True,
            )
            return True
        except CommandError:
            return False

    def in_archive(self, archive_path, file_path):

        if not self.exists(archive_path):
            return False

        loc = str(file_path)
        # query 7z for the specific object location, keeps the output
        # lean, even for big archives
        cmd = f'7z l {sh_quote(str(archive_path))} {sh_quote(loc)}'

        # Note: Currently relies on file_path not showing up in case of failure
        # including non-existent archive. If need be could be more sophisticated
        # and called with check=True + catch RemoteCommandFailedError
        out = self.servershell(
            cmd,
            check=False,
        ).stdout.decode()

        return loc in out

    def get_from_archive(self, archive, src, dst, progress_cb):
        # Note, that as we are in blocking mode, we can't easily fail on the
        # actual get (that is 'cat'). Therefore check beforehand.
        if not self.exists(archive):
            raise RIARemoteError("archive {arc} does not exist."
                                 "".format(arc=archive))

        # with `7z -slt` we get an info block per file like this
        #
        #    Path = some.txt
        #    Size = 4
        #    Packed Size = 8
        #    Modified = 2024-04-18 14:55:39.2376272
        #    Attributes = A -rw-rw-r--
        #    CRC = 5A82FD08
        #    Encrypted = -
        #    Method = LZMA2:12
        #    Block = 0
        #
        # we use -scsUTF-8 to be able to match an UTF filename properly,
        # and otherwise use basic grep/cut to get the integer byte size of
        # the file to be extracted
        #
        size_cmd = \
            f'7z -slt -scsUTF-8 l "{archive}" | grep -A9 "Path = {src}" ' \
            '| grep "^Size =" | cut -d " " -f 3'
        res = self.servershell(size_cmd, check=True)
        nbytes = res.stdout.strip().decode()
        if not nbytes:
            raise RIARemoteError(
                'Cannot determine archive member size. Invalid name?')
        member_size = int(res.stdout.strip().decode())

        cmd = f'7z x -so -- {sh_quote(str(archive))} {sh_quote(str(src))}'
        resgen = self.servershell.start(
            cmd,
            response_generator=FixedLengthResponseGeneratorPosix(
                self.servershell.stdout,
                member_size,
            ),
        )
        bytes_received = 0
        with open(dst, 'wb') as target_file:
            for chunk in resgen:
                bytes_received += len(chunk)
                target_file.write(chunk)
                progress_cb(bytes_received)
            assert resgen.returncode == 0
        if member_size:
            assert member_size == bytes_received

    def read_file(self, file_path):
        cmd = f"cat {sh_quote(str(file_path))}"
        try:
            out = self.servershell(
                cmd,
                check=True,
            ).stdout.decode()
        except CommandError as e:
            # Currently we don't read stderr. All we know is, we couldn't read.
            # Try narrowing it down by calling a subsequent exists()
            if not self.exists(file_path):
                raise FileNotFoundError(f"{str(file_path)} not found.") from e
            else:
                raise RuntimeError(f"Could not read {file_path}") from e

        return out

    def write_file(self, file_path, content, mode='w'):

        if mode == 'w':
            mode = ">"
        elif mode == 'a':
            mode = ">>"
        else:
            raise ValueError("Unknown mode '{}'".format(mode))

        # it really should read from stdin, but MIH cannot make it happen
        stdin = content.encode()
        cmd = f"head -c {len(stdin)} | cat {mode} {sh_quote(str(file_path))}"
        try:
            self.servershell(
                cmd,
                check=True,
                stdin=[stdin],
            )
        except CommandError as e:
            raise RIARemoteError(f"Could not write to {file_path}") from e

    def get_7z(self):
        # TODO: To not rely on availability in PATH we might want to use `which`
        #       (`where` on windows) and get the actual path to 7z to reuse in
        #       in_archive() and get().
        #       Note: `command -v XXX` or `type` might be cross-platform
        #       solution!
        #       However, for availability probing only, it would be sufficient
        #       to just call 7z and see whether it returns zero.

        try:
            self.servershell(
                "7z",
                check=True,
            )
            return True
        except CommandError:
            return False


def oraremote_close_io_onclose(self):
    if self._io:
        self._io.close()
        self._io = None
    if self._push_io:
        self._push_io.close()
        self._push_io = None


# replace the whole class
apply_patch('datalad.distributed.ora_remote', None, 'SSHRemoteIO', SSHRemoteIO)
# add close handler that calls the io.close()
apply_patch('datalad.distributed.ora_remote', 'ORARemote', 'close',
            oraremote_close_io_onclose)



"""Various iterators, e.g., for subprocess pipelining and output processing

.. deprecated:: 1.6
   This module is deprecated. It has been migrated to the `datasalad library
   <https://pypi.org/project/datasalad>`__. Imports should be adjusted to
   ``datasalad.itertools``.
"""

__all__ = [
    'align_pattern',
    'decode_bytes',
    'itemize',
    'load_json',
    'load_json_with_flag',
    'StoreOnly',
    'route_in',
    'route_out',
]

import warnings

from datasalad.itertools import (
    StoreOnly,
    align_pattern,
    decode_bytes,
    itemize,
    load_json,
    load_json_with_flag,
    route_in,
    route_out,
)

warnings.warn(
    '`datalad_next.itertools` has been migrated to the datasalad library, '
    'adjust imports to `datasalad.itertools`',
    DeprecationWarning,
    stacklevel=1,
)



"""Improved credential handling for ``create_sibling_<github-like>()``

This patch makes the storage of a newly entered credential conditional on a
successful authorization, in the spirit of `datalad/datalad#3126
<https://github.com/datalad/datalad/issues/3126>`__.

Moreover, stored credentials now contain a ``realm`` property that
identified the API endpoint. This makes it possible to identify
candidates of suitable credentials without having to specific
their name, similar to a request context url used by the old
providers setup.

This automatic realm-based credential lookup is now also implemented.
When no credential name is specified, the most recently used
credential matching the API realm will be used automatically.
If determined like this, it will be tested for successful
authorization, and will then be stored again with an updated
``last-used`` timestamp.
"""

import logging
from urllib.parse import urlparse

from datalad.distributed.create_sibling_ghlike import _GitHubLike
from datalad.downloaders.http import DEFAULT_USER_AGENT
from datalad_next.exceptions import CapturedException
from datalad_next.utils import CredentialManager

# use same logger as -core
lgr = logging.getLogger('datalad.distributed.create_sibling_ghlike')


def _set_request_headers(self, credential_name, auth_info, require_token):
    credman = CredentialManager()
    from_query = False
    credential = None
    if not credential_name:
        # get the most recent credential by realm, because none was identified
        creds = credman.query(realm=self.api_url, _sortby='last-used')
        if creds:
            # found one, also assign the name to be able to update
            # it below
            credential_name, credential = creds[0]
            from_query = True
    if not credential_name:
        # if we have no name given, fall back on a generated one
        # that may exist from times before realms were recorded
        # properly, otherwise we would not be here
        credential_name = urlparse(self.api_url).netloc
    if not credential:
        # no credential yet
        try:
            credential = credman.get(
                credential_name,
                _prompt=auth_info,
                type='token',
                realm=self.api_url,
            )
            if credential is None or 'secret' not in credential:
                raise ValueError('No credential found')
        except Exception as e:
            CapturedException(e)
            lgr.debug('Token retrieval failed: %s', e)
            lgr.warning(
                'Cannot determine authorization token for %s', credential_name)
            if require_token:
                raise ValueError(
                    f'Authorization required for {self.fullname}, '
                    f'cannot find token for a credential {credential_name}.')
            else:
                lgr.warning("No token found for credential '%s'",
                            credential_name)
            credential = {}

    self.request_headers = {
        'user-agent': DEFAULT_USER_AGENT,
        'authorization':
        f'token {credential.get("secret", "NO-TOKEN-AVAILABLE")}',
    }

    edited_credential = credential.pop('_edited', False)
    if from_query or edited_credential:
        # if the credential was determined based on the api realm or edited,
        # test it so we know it (still) works before we save/update it
        try:
            self.authenticated_user
        except Exception as e:
            raise ValueError(
                ("{state} credential {name!r} did not yield successful "
                 "authorization. {advice}").format(
                    state='Entered' if edited_credential else "Auto-selected",
                    name=credential_name,
                    advice='Please try again with a valid credential'
                    if edited_credential
                    else 'Please select a different credential via the '
                    '`credential` option of this command, '
                    'or remove/edit the credential with the DataLad command '
                    '`credentials`.'
                )
            ) from e
        # this went well, store
        try:
            credman.set(
                credential_name,
                _lastused=True,
                **credential,
            )
        except Exception as e:
            # we do not want to crash for any failure to store a
            # credential
            lgr.warn(
                'Exception raised when storing credential %r %r: %s',
                credential_name,
                credential,
                CapturedException(e),
            )

# patch the core class
lgr.debug('Apply datalad-next patch to create_sibling_ghlike.py:_GitHubLike._set_request_headers')
_GitHubLike._set_request_headers = _set_request_headers

# update docs
_GitHubLike.create_sibling_params['credential']._doc = """\
name of the credential providing a personal access token
to be used for authorization. The token can be supplied via
configuration setting 'datalad.credential.<name>.secret', or
environment variable DATALAD_CREDENTIAL_<NAME>_SECRET, or will
be queried from the active credential store using the provided
name. If none is provided, the last-used token for the
API URL realm will be used. If no matching credential exists,
a credential named after the hostname part of the API URL is tried
as a last fallback."""




"""Common constants

COPY_BUFSIZE
  ``shutil`` buffer size default, with Windows platform default changes
  backported from Python 3.10.

PRE_INIT_COMMIT_SHA
  SHA value for ``git hash-object -t tree /dev/null``, i.e. for nothing.
  This corresponds to the state of a Git repository before the first commit
  is made.

on_linux
  ``True`` if executed on the Linux platform.

on_windows
  ``True`` if executed on the Windows platform.
"""

__all__ = [
    'COPY_BUFSIZE',
    'PRE_INIT_COMMIT_SHA',
    'on_linux',
    'on_windows',
]

# import from "utils", but these really are constants
from datalad.utils import (
    on_linux,
    on_windows,
)

try:
    from shutil import COPY_BUFSIZE
except ImportError:  # pragma: no cover
    # too old
    # from PY3.10
    COPY_BUFSIZE = 1024 * 1024 if on_windows else 64 * 1024

from datalad_core.consts import PRE_INIT_COMMIT_SHA
